#!/usr/bin/env python3

# This app records performance measurements for icechunk reads/writes orchestrated using plain tasks.
# See array.py that orchestrates similar operations using an array library.

# Orchestration
# --executor=threads, --executor=processes, with --num-workers
# --executor=dask-threads, --executor=dask-processes, --executor=dask-distributed --num-workers --threads-per-worker
# --framework=dask-array or --framework=tasks

#
import math
import os
import time
from concurrent import futures
from dataclasses import dataclass
from datetime import timedelta
from enum import StrEnum, auto
from typing import cast
from urllib.parse import urlparse

import numpy as np
import tqdm

# import typer
import zarr
from benchmarks import lib
from icechunk import ForkSession, Repository, S3Options, Storage

BRANCH_NAME = "main"


@dataclass
class Task:
    """A read/write task"""

    # The worker will use this Icechunk store to read/write to the dataset
    session: ForkSession
    # Region of the array to write to
    region: tuple[slice, ...]


@dataclass
class TaskResult:
    session: ForkSession
    time: timedelta


def generate_task_array(task: Task, shape: tuple[int, ...]) -> np.typing.ArrayLike:
    """Generates a random array with the given shape and using the seed in the Task"""
    seed = math.prod(slicer.stop for slicer in task.region)
    np.random.seed(seed)
    return np.random.rand(*shape)


def execute_write_task(task: Task) -> TaskResult:
    """Execute task as a write task.

    This will read the time coordinade from `task` and write a "pancake" in that position,
    using random data. Random data is generated using the task seed.

    Returns the Icechunk store after the write is done.

    As you can see Icechunk stores can be passed to remote workers, and returned from them.
    The reason to return the store is that we'll need all the remote stores, when they are
    done, to be able to do a single, global commit to Icechunk.
    """

    store = task.session.store
    group = zarr.group(store=store, overwrite=False)
    array = cast(zarr.Array, group["array"])
    data = generate_task_array(task, array[task.region].shape)
    assert len(task.region) == array.ndim
    tic = time.perf_counter()
    array[task.region] = data
    toc = time.perf_counter()
    return TaskResult(task.session, time=toc - tic)


def get_storage(url: str) -> Storage:
    url = urlparse(url)
    if url.netloc == "" or url.params != "" or url.query != "" or url.fragment != "":
        raise ValueError(f"Invalid {url=!r}")

    if url.scheme == "s3":
        if url.path == "":
            raise ValueError(f"Invalid {url=!r}")
        storage = Storage.new_s3(bucket=url.netloc, prefix=url.path, config=S3Options())

    elif url.scheme == "local":
        storage = Storage.new_local_filesystem(
            os.path.expanduser(f"{url.netloc}/{url.path}")
        )

    return storage


class Executor(StrEnum):
    threads = auto()
    processes = auto()
    # dask_threads = auto()
    # dask_processes = auto()
    dask_distributed = auto()


# app = typer.Typer()


# @app.command()
def read(
    # data
    url,
    # execution
    executor: Executor = Executor.threads,
    workers: int | None = None,
    threads_per_worker: int | None = None,
) -> None:
    pass


def validate_data(*, num_arrays, shape, chunks) -> None:
    if len(chunks) != len(shape):
        raise ValueError(f"{shape=!r} does not match {chunks=!r}")


def get_executor(
    *,
    executor: Executor,
    workers: int | None = None,
    threads_per_worker: int | None = None,
) -> futures.Executor:
    if executor is Executor.threads:
        return futures.ThreadPoolExecutor(workers)
    elif executor is Executor.processes:
        return futures.ProcessPoolExecutor(workers)
    elif executor is Executor.dask_distributed:
        from distributed import Client

        return Client(n_workers=workers, threads_per_worker=threads_per_worker)
    else:
        raise NotImplementedError


def initialize_store(repo: Repository, *, shape, chunks):
    session = repo.writable_session(BRANCH_NAME)
    store = session.store
    group = zarr.group(store, overwrite=True)
    group.create_array(
        "array", shape=shape, dtype=np.float32, chunks=chunks, overwrite=True
    )
    session.commit("initialized")


# @app.command()
def write(
    # data
    url: str | None = None,
    storage: Storage | None = None,
    # num_arrays: Annotated[int, typer.Option(min=1, max=1)] = 1,
    num_arrays: int = 1,
    # scale:   # TODO: [small, medium, large]
    shape: list[int] = [3, 720, 1440],  # noqa: B006
    chunks: list[int] = [1, -1, -1],  # noqa: B006
    task_nchunks: int = 1,
    # execution
    executor: Executor = Executor.threads,
    workers: int | None = None,
    threads_per_worker: int | None = None,
    record: bool = False,
) -> None:
    timer = lib.Timer()
    validate_data(num_arrays=num_arrays, shape=shape, chunks=chunks)
    chunks = lib.normalize_chunks(shape=shape, chunks=chunks)
    executor = get_executor(
        executor=executor, workers=workers, threads_per_worker=threads_per_worker
    )
    if storage is None:
        storage = get_storage(url)
    repo = Repository.open_or_create(storage)
    initialize_store(repo, shape=shape, chunks=chunks)

    session = repo.writable_session("main")

    task_chunk_shape = lib.get_task_chunk_shape(
        task_nchunks=task_nchunks, shape=shape, chunks=chunks
    )

    # TODO: record nchunks, chunk size in bytes for all timers?
    # TODO: record commit hash of benchmarked commit too.
    timer_context = {}
    with timer.time(op="total_write_time", **timer_context):
        # tasks =
        # print(f"submitted {len(tasks)} tasks")
        fork = session.fork()
        results = [
            f.result(timeout=5)
            for f in tqdm.tqdm(
                futures.as_completed(
                    [
                        executor.submit(
                            execute_write_task, Task(session=fork, region=slicer)
                        )
                        for slicer in lib.slices_from_chunks(shape, task_chunk_shape)
                    ]
                )
            )
        ]

    timer.diagnostics.append(
        {"op": "write_task", "runtime": tuple(r.time for r in results)}
    )

    # TODO: record time & byte size?
    with timer.time(op="merge_changeset", **timer_context):
        if not isinstance(executor, futures.ThreadPoolExecutor):
            session.merge(*(res.session for res in results))
    assert session.has_uncommitted_changes

    with timer.time(op="commit", **timer_context):
        commit_res = session.commit("committed.")
    assert commit_res

    print("Written and committed!")
    return timer


# @app.command()
def verify(
    # data
    # execution
    executor: Executor = Executor.threads,
    workers: int | None = None,
    threads_per_worker: int | None = None,
) -> None:
    pass


# @app.command()
def create(
    # data
    # execution
    executor: Executor = Executor.threads,
    workers: int | None = None,
    threads_per_worker: int | None = None,
) -> None:
    pass


# if __name__ == "__main__":
#     app()



from __future__ import annotations

import operator
from typing import Literal, cast

import numpy as np
import pytest

import zarr
import zarr.api
import zarr.api.asynchronous
from icechunk import IcechunkError, IcechunkStore
from tests.conftest import parse_repo
from zarr import Array, AsyncGroup, Group
from zarr.api.asynchronous import create_array
from zarr.core.buffer import default_buffer_prototype
from zarr.core.common import JSON, ZarrFormat
from zarr.core.group import GroupMetadata
from zarr.core.sync import sync
from zarr.errors import ContainsArrayError, ContainsGroupError
from zarr.storage import StorePath
from zarr.storage._common import make_store_path


@pytest.fixture(params=["memory"])
def store(request: pytest.FixtureRequest) -> IcechunkStore:
    repo = parse_repo(request.param, "test")
    session = repo.writable_session("main")
    return session.store


@pytest.fixture(params=[True, False])
def overwrite(request: pytest.FixtureRequest) -> bool:
    result = request.param
    if not isinstance(result, bool):
        raise TypeError("Wrong type returned by test fixture.")
    return result


@pytest.fixture(params=[3], ids=["zarr3"])
def zarr_format(request: pytest.FixtureRequest) -> ZarrFormat:
    result = request.param
    if result not in [3]:
        raise ValueError("Wrong value returned from test fixture.")
    return cast(ZarrFormat, result)


async def test_create_creates_parents(
    store: IcechunkStore, zarr_format: ZarrFormat
) -> None:
    # prepare a root node, with some data set
    await zarr.api.asynchronous.open_group(
        store=store, path="a", zarr_format=zarr_format, attributes={"key": "value"}
    )
    objs = {x async for x in store.list()}
    if zarr_format == 2:
        assert objs == {".zgroup", ".zattrs", "a/.zgroup", "a/.zattrs"}
    else:
        assert objs == {"zarr.json", "a/zarr.json"}

    # test that root group node was created
    root = await zarr.api.asynchronous.open_group(
        store=store,
    )

    agroup = await root.getitem("a")
    assert agroup.attrs == {"key": "value"}

    # create a child node with a couple intermediates
    await zarr.api.asynchronous.open_group(
        store=store, path="a/b/c/d", zarr_format=zarr_format
    )
    parts = ["a", "a/b", "a/b/c"]

    if zarr_format == 2:
        files = [".zattrs", ".zgroup"]
    else:
        files = ["zarr.json"]

    expected = [f"{part}/{file}" for file in files for part in parts]

    if zarr_format == 2:
        expected.extend([".zgroup", ".zattrs", "a/b/c/d/.zgroup", "a/b/c/d/.zattrs"])
    else:
        expected.extend(["zarr.json", "a/b/c/d/zarr.json"])

    expected = sorted(expected)

    result = sorted([x async for x in store.list_prefix("")])

    assert result == expected

    paths = ["a", "a/b", "a/b/c"]
    for path in paths:
        g = await zarr.api.asynchronous.open_group(store=store, path=path)
        assert isinstance(g, AsyncGroup)

        if path == "a":
            # ensure we didn't overwrite the root attributes
            assert g.attrs == {"key": "value"}
        else:
            assert g.attrs == {}


def test_group_members(store: IcechunkStore, zarr_format: ZarrFormat) -> None:
    """
    Test that `Group.members` returns correct values, i.e. the arrays and groups
    (explicit and implicit) contained in that group.
    """

    path = "group"
    group = Group.from_store(store=store, zarr_format=zarr_format)
    members_expected: dict[str, Array | Group] = {}

    members_expected["subgroup"] = group.create_group("subgroup")
    # make a sub-sub-subgroup, to ensure that the children calculation doesn't go
    # too deep in the hierarchy
    subsubgroup = cast(Group, members_expected["subgroup"]).create_group("subsubgroup")
    subsubsubgroup = subsubgroup.create_group("subsubsubgroup")

    members_expected["subarray"] = group.create_array(
        "subarray", shape=(100,), dtype="uint8", chunks=(10,), overwrite=True
    )

    # This is not supported by Icechunk, so we expect an error
    # zarr-python: add an extra object to the domain of the group.
    # the list of children should ignore this object.
    with pytest.raises(IcechunkError):
        sync(
            store.set(
                f"{path}/extra_object-1",
                default_buffer_prototype().buffer.from_bytes(b"000000"),
            )
        )

    # This is not supported by Icechunk, so we expect an error
    # zarr-python: add an extra object under a directory-like prefix in the domain of the group.
    # this creates a directory with a random key in it
    # this should not show up as a member
    with pytest.raises(IcechunkError):
        sync(
            store.set(
                f"{path}/extra_directory/extra_object-2",
                default_buffer_prototype().buffer.from_bytes(b"000000"),
            )
        )
    members_observed = group.members()
    # members are not guaranteed to be ordered, so sort before comparing
    assert sorted(dict(members_observed)) == sorted(members_expected)

    # partial
    members_observed = group.members(max_depth=1)
    members_expected["subgroup/subsubgroup"] = subsubgroup
    # members are not guaranteed to be ordered, so sort before comparing
    assert sorted(dict(members_observed)) == sorted(members_expected)

    # total
    members_observed = group.members(max_depth=None)
    members_expected["subgroup/subsubgroup/subsubsubgroup"] = subsubsubgroup
    # members are not guaranteed to be ordered, so sort before comparing
    assert sorted(dict(members_observed)) == sorted(members_expected)

    with pytest.raises(ValueError, match="max_depth"):
        members_observed = group.members(max_depth=-1)


def test_group(store: IcechunkStore, zarr_format: ZarrFormat) -> None:
    """
    Test basic Group routines.
    """
    store_path = StorePath(store)
    agroup = AsyncGroup(
        metadata=GroupMetadata(zarr_format=zarr_format), store_path=store_path
    )
    group = Group(agroup)
    assert agroup.metadata is group.metadata
    assert agroup.store_path == group.store_path == store_path

    # create two groups
    foo = group.create_group("foo")
    bar = foo.create_group("bar", attributes={"baz": "qux"})

    # create an array from the "bar" group
    data = np.arange(0, 4 * 4, dtype="uint16").reshape((4, 4))
    arr = bar.create_array(
        "baz", shape=data.shape, dtype=data.dtype, chunks=(2, 2), overwrite=True
    )
    arr[:] = data

    # check the array
    assert arr == bar["baz"]
    assert arr.shape == data.shape
    assert arr.dtype == data.dtype

    # TODO: update this once the array api settles down
    assert arr.chunks == (2, 2)

    bar2 = foo["bar"]
    assert dict(bar2.attrs) == {"baz": "qux"}

    # update a group's attributes
    bar2.attrs.update({"name": "bar"})
    # bar.attrs was modified in-place
    assert dict(bar2.attrs) == {"baz": "qux", "name": "bar"}

    # and the attrs were modified in the store
    bar3 = foo["bar"]
    assert dict(bar3.attrs) == {"baz": "qux", "name": "bar"}


def test_group_create(
    store: IcechunkStore, overwrite: bool, zarr_format: ZarrFormat
) -> None:
    """
    Test that `Group.create` works as expected.
    """
    attributes = {"foo": 100}
    group = Group.from_store(
        store, attributes=attributes, zarr_format=zarr_format, overwrite=overwrite
    )

    assert group.attrs == attributes

    if not overwrite:
        with pytest.raises(ContainsGroupError):
            _ = Group.from_store(store, overwrite=overwrite, zarr_format=zarr_format)


def test_group_open(
    store: IcechunkStore, zarr_format: ZarrFormat, overwrite: bool
) -> None:
    """
    Test the `Group.open` method.
    """
    spath = StorePath(store)
    # attempt to open a group that does not exist
    with pytest.raises(FileNotFoundError):
        Group.open(store)

    # create the group
    attrs = {"path": "foo"}
    group_created = Group.from_store(
        store, attributes=attrs, zarr_format=zarr_format, overwrite=overwrite
    )
    assert group_created.attrs == attrs
    assert group_created.metadata.zarr_format == zarr_format
    assert group_created.store_path == spath

    # attempt to create a new group in place, to test overwrite
    new_attrs = {"path": "bar"}
    if not overwrite:
        with pytest.raises(ContainsGroupError):
            Group.from_store(
                store, attributes=attrs, zarr_format=zarr_format, overwrite=overwrite
            )
    else:
        if not store.supports_deletes:
            pytest.skip(
                "Store does not support deletes but `overwrite` is True, requiring deletes to override a group"
            )
        group_created_again = Group.from_store(
            store, attributes=new_attrs, zarr_format=zarr_format, overwrite=overwrite
        )
        assert group_created_again.attrs == new_attrs
        assert group_created_again.metadata.zarr_format == zarr_format
        assert group_created_again.store_path == spath


def test_group_getitem(store: IcechunkStore, zarr_format: ZarrFormat) -> None:
    """
    Test the `Group.__getitem__` method.
    """

    group = Group.from_store(store, zarr_format=zarr_format)
    subgroup = group.create_group(name="subgroup")
    subarray = group.create_array(
        name="subarray", shape=(10,), chunks=(10,), dtype="uint8"
    )

    assert group["subgroup"] == subgroup
    assert group["subarray"] == subarray
    with pytest.raises(KeyError):
        group["nope"]


def test_group_delitem(store: IcechunkStore, zarr_format: ZarrFormat) -> None:
    """
    Test the `Group.__delitem__` method.
    """

    group = Group.from_store(store, zarr_format=zarr_format)
    subgroup = group.create_group(name="subgroup")
    subarray = group.create_array(
        name="subarray", shape=(10,), chunks=(10,), dtype="uint8"
    )

    assert group["subgroup"] == subgroup
    assert group["subarray"] == subarray

    del group["subgroup"]
    with pytest.raises(KeyError):
        group["subgroup"]

    del group["subarray"]
    with pytest.raises(KeyError):
        group["subarray"]


def test_group_setitem(store: IcechunkStore, zarr_format: ZarrFormat) -> None:
    """
    Test the `Group.__setitem__` method.
    """
    group = Group.from_store(store, zarr_format=zarr_format)
    arr = np.ones((2, 4))
    group["key"] = arr
    assert list(group.array_keys()) == ["key"]
    assert isinstance(group["key"], Array)
    assert group["key"].shape == (2, 4)
    np.testing.assert_array_equal(group["key"][:], arr)

    if store.supports_deletes:
        key = "key"
    else:
        # overwriting with another array requires deletes
        # for stores that don't support this, we just use a new key
        key = "key2"

        # overwrite with another array
        arr = np.zeros((3, 5))
        group[key] = arr
        assert key in list(group.array_keys())
        a = group[key]
        assert isinstance(a, Array)
        assert a.shape == (3, 5)
        np.testing.assert_array_equal(a, arr)  # type: ignore [arg-type]


def test_group_contains(store: IcechunkStore, zarr_format: ZarrFormat) -> None:
    """
    Test the `Group.__contains__` method
    """
    group = Group.from_store(store, zarr_format=zarr_format)
    assert "foo" not in group
    _ = group.create_group(name="foo")
    assert "foo" in group


def test_group_subgroups(store: IcechunkStore, zarr_format: ZarrFormat) -> None:
    """
    Test the behavior of `Group` methods for accessing subgroups, namely `Group.group_keys` and `Group.groups`
    """
    group = Group.from_store(store, zarr_format=zarr_format)
    keys = ("foo", "bar")
    subgroups_expected = tuple(group.create_group(k) for k in keys)
    # create a sub-array as well
    _ = group.create_array("array", shape=(10,), dtype="uint8")
    subgroups_observed = tuple(a[1] for a in group.groups())
    assert set(group.group_keys()) == set(keys)
    assert len(subgroups_observed) == len(subgroups_expected)
    assert all(a in subgroups_observed for a in subgroups_expected)


def test_group_subarrays(store: IcechunkStore, zarr_format: ZarrFormat) -> None:
    """
    Test the behavior of `Group` methods for accessing subgroups, namely `Group.group_keys` and `Group.groups`
    """
    group = Group.from_store(store, zarr_format=zarr_format)
    keys = ("foo", "bar")
    subarrays_expected = tuple(
        group.create_array(k, shape=(10,), dtype="uint8") for k in keys
    )
    # create a sub-group as well
    _ = group.create_group("group")
    subarrays_observed = tuple(a[1] for a in group.arrays())
    assert set(group.array_keys()) == set(keys)
    assert len(subarrays_observed) == len(subarrays_expected)
    assert all(a in subarrays_observed for a in subarrays_expected)


def test_group_update_attributes(store: IcechunkStore, zarr_format: ZarrFormat) -> None:
    """
    Test the behavior of `Group.update_attributes`
    """
    attrs = {"foo": 100}
    group = Group.from_store(store, zarr_format=zarr_format, attributes=attrs)
    assert group.attrs == attrs
    new_attrs = {"bar": 100}
    new_group = group.update_attributes(new_attrs)
    updated_attrs = attrs.copy()
    updated_attrs.update(new_attrs)
    assert new_group.attrs == updated_attrs


async def test_group_update_attributes_async(
    store: IcechunkStore, zarr_format: ZarrFormat
) -> None:
    """
    Test the behavior of `Group.update_attributes_async`
    """
    attrs = {"foo": 100}
    group = Group.from_store(store, zarr_format=zarr_format, attributes=attrs)
    assert group.attrs == attrs
    new_attrs = {"bar": 100}
    new_group = await group.update_attributes_async(new_attrs)
    assert new_group.attrs == new_attrs


def test_group_create_array(
    store: IcechunkStore,
    zarr_format: ZarrFormat,
    overwrite: bool,
) -> None:
    """
    Test `Group.create_array`
    """
    group = Group.from_store(store, zarr_format=zarr_format)
    shape = (10, 10)
    dtype = "uint8"
    data = np.arange(np.prod(shape)).reshape(shape).astype(dtype)

    array = group.create_array(name="array", shape=shape, dtype=dtype)
    array[:] = data

    if not overwrite:
        with pytest.raises(ContainsArrayError):
            array = group.create_array(name="array", shape=shape, dtype=dtype)
            array[:] = data
    assert array.shape == shape
    assert array.dtype == np.dtype(dtype)
    assert np.array_equal(array[:], data)


@pytest.mark.parametrize("store", ["memory"], indirect=["store"])
@pytest.mark.parametrize("zarr_format", [3])
@pytest.mark.parametrize("overwrite", [True, False])
@pytest.mark.parametrize("extant_node", ["array", "group"])
def test_group_creation_existing_node(
    store: IcechunkStore,
    zarr_format: ZarrFormat,
    overwrite: bool,
    extant_node: Literal["array", "group"],
) -> None:
    """
    Check that an existing array or group is handled as expected during group creation.
    """
    spath = StorePath(store)
    group = Group.from_store(spath, zarr_format=zarr_format)
    expected_exception: type[ContainsArrayError | ContainsGroupError]
    attributes: dict[str, JSON] = {"old": True}

    if extant_node == "array":
        expected_exception = ContainsArrayError
        _ = group.create_array(
            "extant", shape=(10,), dtype="uint8", attributes=attributes
        )
    elif extant_node == "group":
        expected_exception = ContainsGroupError
        _ = group.create_group("extant", attributes=attributes)
    else:
        raise AssertionError

    new_attributes = {"new": True}

    if overwrite:
        node_new = Group.from_store(
            spath / "extant",
            attributes=new_attributes,
            zarr_format=zarr_format,
            overwrite=overwrite,
        )
        assert node_new.attrs == new_attributes
    else:
        with pytest.raises(expected_exception):
            Group.from_store(
                spath / "extant",
                attributes=new_attributes,
                zarr_format=zarr_format,
                overwrite=overwrite,
            )


async def test_asyncgroup_create(
    store: IcechunkStore,
    overwrite: bool,
    zarr_format: ZarrFormat,
) -> None:
    """
    Test that `AsyncGroup.from_store` works as expected.
    """
    spath = StorePath(store=store)
    attributes = {"foo": 100}
    agroup = await AsyncGroup.from_store(
        store,
        attributes=attributes,
        overwrite=overwrite,
        zarr_format=zarr_format,
    )

    assert agroup.metadata == GroupMetadata(
        zarr_format=zarr_format, attributes=attributes
    )
    assert agroup.store_path == await make_store_path(store)

    if not overwrite:
        with pytest.raises(ContainsGroupError):
            agroup = await AsyncGroup.from_store(
                spath,
                attributes=attributes,
                overwrite=overwrite,
                zarr_format=zarr_format,
            )
        # create an array at our target path
        collision_name = "foo"
        _ = await create_array(
            spath / collision_name, shape=(10,), dtype="uint8", zarr_format=zarr_format
        )
        with pytest.raises(ContainsArrayError):
            _ = await AsyncGroup.from_store(
                StorePath(store=store) / collision_name,
                attributes=attributes,
                overwrite=overwrite,
                zarr_format=zarr_format,
            )


# todo: replace this with a declarative API where we model a full hierarchy


async def test_asyncgroup_delitem(store: IcechunkStore, zarr_format: ZarrFormat) -> None:
    agroup = await AsyncGroup.from_store(store=store, zarr_format=zarr_format)
    array_name = "sub_array"
    _ = await agroup.create_array(
        name=array_name,
        shape=(10,),
        dtype="uint8",
        chunks=(2,),
        attributes={"foo": 100},
    )
    await agroup.delitem(array_name)

    #  todo: clean up the code duplication here
    if zarr_format == 2:
        assert not await agroup.store_path.store.exists(array_name + "/" + ".zarray")
        assert not await agroup.store_path.store.exists(array_name + "/" + ".zattrs")
    elif zarr_format == 3:
        assert not await agroup.store_path.store.exists(array_name + "/" + "zarr.json")
    else:
        raise AssertionError

    sub_group_path = "sub_group"
    _ = await agroup.create_group(sub_group_path, attributes={"foo": 100})
    await agroup.delitem(sub_group_path)
    if zarr_format == 2:
        assert not await agroup.store_path.store.exists(array_name + "/" + ".zgroup")
        assert not await agroup.store_path.store.exists(array_name + "/" + ".zattrs")
    elif zarr_format == 3:
        assert not await agroup.store_path.store.exists(array_name + "/" + "zarr.json")
    else:
        raise AssertionError


async def test_asyncgroup_update_attributes(
    store: IcechunkStore, zarr_format: ZarrFormat
) -> None:
    """
    Test that the AsyncGroup.update_attributes method works correctly.
    """
    attributes_old = {"foo": 10}
    attributes_new = {"baz": "new"}
    agroup = await AsyncGroup.from_store(
        store=store, zarr_format=zarr_format, attributes=attributes_old
    )

    agroup_new_attributes = await agroup.update_attributes(attributes_new)
    assert agroup_new_attributes.attrs == attributes_old | attributes_new


async def test_group_members_async(store: IcechunkStore) -> None:
    group = AsyncGroup(
        GroupMetadata(),
        store_path=StorePath(store=store, path="root"),
    )
    a0 = await group.create_array("a0", shape=(1,), dtype="uint8")
    g0 = await group.create_group("g0")
    a1 = await g0.create_array("a1", shape=(1,), dtype="uint8")
    g1 = await g0.create_group("g1")
    a2 = await g1.create_array("a2", shape=(1,), dtype="uint8")
    g2 = await g1.create_group("g2")

    # immediate children
    children = sorted([x async for x in group.members()], key=operator.itemgetter(0))
    assert children == [
        ("a0", a0),
        ("g0", g0),
    ]

    nmembers = await group.nmembers()
    assert nmembers == 2

    # partial
    children = sorted(
        [x async for x in group.members(max_depth=1)], key=operator.itemgetter(0)
    )
    expected = [
        ("a0", a0),
        ("g0", g0),
        ("g0/a1", a1),
        ("g0/g1", g1),
    ]
    assert children == expected
    nmembers = await group.nmembers(max_depth=1)
    assert nmembers == 4

    # all children
    all_children = sorted(
        [x async for x in group.members(max_depth=None)], key=operator.itemgetter(0)
    )
    expected = [
        ("a0", a0),
        ("g0", g0),
        ("g0/a1", a1),
        ("g0/g1", g1),
        ("g0/g1/a2", a2),
        ("g0/g1/g2", g2),
    ]
    assert all_children == expected

    nmembers = await group.nmembers(max_depth=None)
    assert nmembers == 6

    with pytest.raises(ValueError, match="max_depth"):
        [x async for x in group.members(max_depth=-1)]


class TestGroupMetadata:
    def test_from_dict_extra_fields(self) -> None:
        data = {
            "attributes": {"key": "value"},
            "_nczarr_superblock": {"version": "2.0.0"},
            "zarr_format": 2,
        }
        result = GroupMetadata.from_dict(data)
        expected = GroupMetadata(attributes={"key": "value"}, zarr_format=2)
        assert result == expected


def test_delitem_removes_children(store: IcechunkStore, zarr_format: ZarrFormat) -> None:
    # https://github.com/zarr-developers/zarr-python/issues/2191
    g1 = zarr.group(store=store, zarr_format=zarr_format)
    g1.create_group("0")
    g1.create_group("0/0")
    arr = g1.create_array("0/0/0", shape=(1,), dtype="uint8")
    arr[:] = 1

    del g1["0"]
    with pytest.raises(KeyError):
        g1["0/0"]



from typing import Literal

import boto3
import pytest
from mypy_boto3_s3.client import S3Client

from icechunk import Repository, in_memory_storage, local_filesystem_storage


def parse_repo(store: Literal["local", "memory"], path: str) -> Repository:
    if store == "local":
        return Repository.create(
            storage=local_filesystem_storage(path),
        )
    if store == "memory":
        return Repository.create(
            storage=in_memory_storage(),
        )


@pytest.fixture(scope="function")
def repo(request: pytest.FixtureRequest, tmpdir: str) -> tuple[Repository, str]:
    param = request.param
    repo = parse_repo(param, tmpdir)
    return repo, tmpdir


minio_client = None


def get_minio_client() -> S3Client:
    global minio_client
    if minio_client is None:
        minio_client = boto3.client(
            "s3",
            endpoint_url="http://localhost:9000",
            use_ssl=False,
            aws_access_key_id="minio123",
            aws_secret_access_key="minio123",
        )
    return minio_client


def write_chunks_to_minio(
    chunks: list[tuple[str, bytes]], bucket: str = "testbucket"
) -> list[str]:
    """Write chunks to local minio returning their etags"""
    s3 = get_minio_client()
    etags = []
    for key, data in chunks:
        etag = s3.put_object(Bucket=bucket, Key=key, Body=data)["ETag"]
        etags.append(etag)

    return etags



import logging
import os
import subprocess

import icechunk as ic


def setup_logger():
    logger = logging.getLogger("icechunk-bench")
    logger.setLevel(logging.INFO)
    console_handler = logging.StreamHandler()
    logger.addHandler(console_handler)
    logger.handlers = logger.handlers[:1]  # make idempotent
    return logger


def get_coiled_kwargs(*, store: str, region: str | None = None) -> str:
    if store == "s3_ob":
        store = "s3"
    COILED_VM_TYPES = {
        # TODO: think about these
        "s3": "m5.4xlarge",
        "gcs": "n2-standard-16",
        "tigris": "m5.4xlarge",
        "r2": "m5.4xlarge",
    }
    DEFAULT_REGIONS = {
        "s3": "us-east-1",
        "gcs": "us-east1",
        "tigris": "us-east-1",
        "r2": "us-west-2",
        "az": "eastus",
    }
    WORKSPACES = {
        "s3": "earthmover-devs",
        "tigris": "earthmover-devs",
        "r2": "earthmover-devs",
        "gcs": "earthmover-devs-gcp",
        "az": "earthmover-devs-azure",
    }
    TIGRIS_REGIONS = {"iad": "us-east-1"}

    if region is None:
        region = DEFAULT_REGIONS[store]
    else:
        region = TIGRIS_REGIONS[region] if store == "tigris" else region
    return {
        "workspace": WORKSPACES[store],
        "region": region,
        "vm_type": COILED_VM_TYPES[store],
    }


def assert_cwd_is_icechunk_python():
    CURRENTDIR = os.getcwd()
    if not CURRENTDIR.endswith("icechunk-python"):
        raise ValueError(
            "Running in the wrong directory. Please run from the `icechunk-python` directory."
        )


def get_full_commit(ref: str) -> str:
    return subprocess.run(
        ["git", "rev-parse", ref], capture_output=True, text=True, check=True
    ).stdout.strip()


def rdms() -> str:
    import random
    import string

    return "".join(random.sample(string.ascii_lowercase, k=8))


def repo_config_with(
    *, inline_chunk_threshold_bytes: int | None = None, preload=None, splitting=None
) -> ic.RepositoryConfig:
    config = ic.RepositoryConfig.default()
    if inline_chunk_threshold_bytes is not None:
        config.inline_chunk_threshold_bytes = inline_chunk_threshold_bytes
    if splitting is not None:
        config.manifest = ic.ManifestConfig(preload=preload, splitting=splitting)
    return config


def get_splitting_config(*, split_size: int):
    # helper to allow benchmarking versions before manifest splitting was introduced
    from icechunk import (
        ManifestSplitCondition,
        ManifestSplitDimCondition,
        ManifestSplittingConfig,
    )

    return ManifestSplittingConfig.from_dict(
        {
            ManifestSplitCondition.path_matches(".*"): {
                ManifestSplitDimCondition.Any(): split_size
            }
        }
    )



# An example of using multiprocessing to write to an Icechunk dataset

import tempfile
from concurrent.futures import ProcessPoolExecutor

import xarray as xr
from icechunk import Repository, Session, local_filesystem_storage


def write_timestamp(*, itime: int, session: Session) -> Session:
    # pass a list to isel to preserve the time dimension
    ds = xr.tutorial.open_dataset("rasm").isel(time=[itime])
    # region="auto" tells Xarray to infer which "region" of the output arrays to write to.
    ds.to_zarr(session.store, region="auto", consolidated=False)
    return session


if __name__ == "__main__":
    ds = xr.tutorial.open_dataset("rasm").isel(time=slice(24))
    repo = Repository.create(local_filesystem_storage(tempfile.mkdtemp()))
    session = repo.writable_session("main")

    chunks = {1 if dim == "time" else ds.sizes[dim] for dim in ds.Tair.dims}
    ds.to_zarr(
        session.store,
        compute=False,
        encoding={"Tair": {"chunks": chunks}},
        mode="w",
        consolidated=False,
    )
    # this commit is optional, but may be useful in your workflow
    session.commit("initialize store")

    session = repo.writable_session("main")
    fork = session.fork()
    with ProcessPoolExecutor() as executor:
        # opt-in to successful pickling of a writable session
        # submit the writes
        futures = [
            executor.submit(write_timestamp, itime=i, session=fork)
            for i in range(ds.sizes["time"])
        ]
        # grab the Session objects from each individual write task
        fork_sessions = [f.result() for f in futures]

    # manually merge the remote sessions in to the local session
    session.merge(*fork_sessions)
    session.commit("finished writes")

    ondisk = xr.open_zarr(repo.readonly_session("main").store, consolidated=False)
    xr.testing.assert_identical(ds, ondisk)



import contextlib
import warnings
from collections.abc import AsyncIterator, Generator
from typing import Any, NoReturn, Self

from icechunk import (
    ConflictSolver,
    Diff,
    RepositoryConfig,
)
from icechunk._icechunk_python import PySession
from icechunk.store import IcechunkStore


class Session:
    """A session object that allows for reading and writing data from an Icechunk repository."""

    _session: PySession
    _allow_changes: bool

    def __init__(self, session: PySession):
        self._session = session
        self._allow_changes = False

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Session):
            return False
        return self._session == value._session

    def __getstate__(self) -> object:
        if not self.read_only:
            raise ValueError(
                "You must opt-in to pickle writable sessions in a distributed context "
                "using Session.fork(). "
                # link to docs
                "If you are using xarray's `Dataset.to_zarr` method to write dask arrays, "
                "please use `icechunk.xarray.to_icechunk` instead. "
            )
        state = {
            "_session": self._session.as_bytes(),
            "_allow_changes": self._allow_changes,
        }
        return state

    def __setstate__(self, state: object) -> None:
        if not isinstance(state, dict):
            raise ValueError("Invalid state")
        self._session = PySession.from_bytes(state["_session"])
        self._allow_changes = state["_allow_changes"]

    @contextlib.contextmanager
    def allow_pickling(self) -> Generator[None, None, None]:
        """
        Context manager to allow unpickling this store if writable.
        """
        raise RuntimeError(
            "The allow_pickling context manager has been removed. "
            "Use the new `Session.fork` API instead. "
            # FIXME: Add link to docs
            "Better yet, use `to_icechunk` if that will fit your needs."
        )

    @property
    def read_only(self) -> bool:
        """
        Whether the session is read-only.

        Returns
        -------
        bool
            True if the session is read-only, False otherwise.
        """
        return self._session.read_only

    @property
    def snapshot_id(self) -> str:
        """
        The base snapshot ID of the session.

        Returns
        -------
        str
            The base snapshot ID of the session.
        """
        return self._session.snapshot_id

    @property
    def branch(self) -> str | None:
        """
        The branch that the session is based on. This is only set if the session is writable.

        Returns
        -------
        str or None
            The branch that the session is based on if the session is writable, None otherwise.
        """
        return self._session.branch

    @property
    def has_uncommitted_changes(self) -> bool:
        """
        Whether the session has uncommitted changes. This is only possibly true if the session is writable.

        Returns
        -------
        bool
            True if the session has uncommitted changes, False otherwise.
        """
        return self._session.has_uncommitted_changes

    def status(self) -> Diff:
        """
        Compute an overview of the current session changes

        Returns
        -------
        Diff
            The operations executed in the current session but still not committed.
        """
        return self._session.status()

    def discard_changes(self) -> None:
        """
        When the session is writable, discard any uncommitted changes.
        """
        self._session.discard_changes()

    @property
    def store(self) -> IcechunkStore:
        """
        Get a zarr Store object for reading and writing data from the repository using zarr python.

        Returns
        -------
        IcechunkStore
            A zarr Store object for reading and writing data from the repository.
        """
        return IcechunkStore(self._session.store, for_fork=False)

    @property
    def config(self) -> RepositoryConfig:
        """
        Get the repository configuration.

        Notice that changes to the returned object won't be impacted. To change configuration values
        use `Repository.reopen`.

        Returns
        -------
        RepositoryConfig
            The config for the repository that owns this session.
        """
        return self._session.config

    def all_virtual_chunk_locations(self) -> list[str]:
        """
        Return the location URLs of all virtual chunks.

        Returns
        -------
        list of str
            The location URLs of all virtual chunks.
        """
        return self._session.all_virtual_chunk_locations()

    async def all_virtual_chunk_locations_async(self) -> list[str]:
        """
        Return the location URLs of all virtual chunks (async version).

        Returns
        -------
        list of str
            The location URLs of all virtual chunks.
        """
        return await self._session.all_virtual_chunk_locations_async()

    async def chunk_coordinates(
        self, array_path: str, batch_size: int = 1000
    ) -> AsyncIterator[tuple[int, ...]]:
        """
        Return an async iterator to all initialized chunks for the array at array_path

        Returns
        -------
        an async iterator to chunk coordinates as tuples
        """
        # We do unbatching here to improve speed. Switching to rust to get
        # a batch is much faster than switching for every element
        async for batch in self._session.chunk_coordinates(array_path, batch_size):
            for coord in batch:
                yield tuple(coord)

    def merge(self, *others: "ForkSession") -> None:
        """
        Merge the changes for this session with the changes from another session.

        Parameters
        ----------
        others : ForkSession
            The forked sessions to merge changes from.
        """
        for other in others:
            if not isinstance(other, ForkSession):
                raise TypeError(
                    "Sessions can only be merged with a ForkSession created with Session.fork(). "
                    f"Received {type(other).__name__} instead."
                )
            self._session.merge(other._session)
        self._allow_changes = False

    async def merge_async(self, *others: "ForkSession") -> None:
        """
        Merge the changes for this session with the changes from another session (async version).

        Parameters
        ----------
        others : ForkSession
            The forked sessions to merge changes from.
        """
        for other in others:
            if not isinstance(other, ForkSession):
                raise TypeError(
                    "Sessions can only be merged with a ForkSession created with Session.fork(). "
                    f"Received {type(other).__name__} instead."
                )
            await self._session.merge_async(other._session)
        self._allow_changes = False

    def commit(
        self,
        message: str,
        metadata: dict[str, Any] | None = None,
        rebase_with: ConflictSolver | None = None,
        rebase_tries: int = 1_000,
    ) -> str:
        """
        Commit the changes in the session to the repository.

        When successful, the writable session is completed and the session is now read-only and based on the new commit. The snapshot ID of the new commit is returned.

        If the session is out of date, this will raise a ConflictError exception depicting the conflict that occurred. The session will need to be rebased before committing.

        Parameters
        ----------
        message : str
            The message to write with the commit.
        metadata : dict[str, Any] | None, optional
            Additional metadata to store with the commit snapshot.
        rebase_with : ConflictSolver | None, optional
            If other session committed while the current session was writing, use Session.rebase with this solver.
        rebase_tries : int, optional
            If other session committed while the current session was writing, use Session.rebase up to this many times in a loop.

        Returns
        -------
        str
            The snapshot ID of the new commit.

        Raises
        ------
        icechunk.ConflictError
            If the session is out of date and a conflict occurs.
        """
        if self._allow_changes:
            warnings.warn(
                "Committing a session after forking, and without merging will not work. "
                "Merge back in the remote changes first using Session.merge().",
                UserWarning,
                stacklevel=2,
            )
        return self._session.commit(
            message, metadata, rebase_with=rebase_with, rebase_tries=rebase_tries
        )

    async def commit_async(
        self,
        message: str,
        metadata: dict[str, Any] | None = None,
        rebase_with: ConflictSolver | None = None,
        rebase_tries: int = 1_000,
    ) -> str:
        """
        Commit the changes in the session to the repository (async version).

        When successful, the writable session is completed and the session is now read-only and based on the new commit. The snapshot ID of the new commit is returned.

        If the session is out of date, this will raise a ConflictError exception depicting the conflict that occurred. The session will need to be rebased before committing.

        Parameters
        ----------
        message : str
            The message to write with the commit.
        metadata : dict[str, Any] | None, optional
            Additional metadata to store with the commit snapshot.
        rebase_with : ConflictSolver | None, optional
            If other session committed while the current session was writing, use Session.rebase with this solver.
        rebase_tries : int, optional
            If other session committed while the current session was writing, use Session.rebase up to this many times in a loop.

        Returns
        -------
        str
            The snapshot ID of the new commit.

        Raises
        ------
        icechunk.ConflictError
            If the session is out of date and a conflict occurs.
        """
        if self._allow_changes:
            warnings.warn(
                "Committing a session after forking, and without merging will not work. "
                "Merge back in the remote changes first using Session.merge().",
                UserWarning,
                stacklevel=2,
            )
        return await self._session.commit_async(
            message, metadata, rebase_with=rebase_with, rebase_tries=rebase_tries
        )

    def rebase(self, solver: ConflictSolver) -> None:
        """
        Rebase the session to the latest ancestry of the branch.

        This method will iteratively crawl the ancestry of the branch and apply the changes from the branch to the session. If a conflict is detected, the conflict solver will be used to optionally resolve the conflict. When complete, the session will be based on the latest commit of the branch and the session will be ready to attempt another commit.

        When a conflict is detected and a resolution is not possible with the provided solver, a RebaseFailed exception will be raised. This exception will contain the snapshot ID that the rebase failed on and a list of conflicts that occurred.

        Parameters
        ----------
        solver : ConflictSolver
            The conflict solver to use when a conflict is detected.

        Raises
        ------
        RebaseFailedError
            When a conflict is detected and the solver fails to resolve it.
        """
        self._session.rebase(solver)

    async def rebase_async(self, solver: ConflictSolver) -> None:
        """
        Rebase the session to the latest ancestry of the branch (async version).

        This method will iteratively crawl the ancestry of the branch and apply the changes from the branch to the session. If a conflict is detected, the conflict solver will be used to optionally resolve the conflict. When complete, the session will be based on the latest commit of the branch and the session will be ready to attempt another commit.

        When a conflict is detected and a resolution is not possible with the provided solver, a RebaseFailed exception will be raised. This exception will contain the snapshot ID that the rebase failed on and a list of conflicts that occurred.

        Parameters
        ----------
        solver : ConflictSolver
            The conflict solver to use when a conflict is detected.

        Raises
        ------
        RebaseFailedError
            When a conflict is detected and the solver fails to resolve it.
        """
        await self._session.rebase_async(solver)

    def fork(self) -> "ForkSession":
        if self.has_uncommitted_changes:
            raise ValueError(
                "Cannot fork a Session with uncommitted changes. "
                "Make a commit, create a new Session, and then fork that to execute distributed writes."
            )
        if self.read_only:
            raise ValueError(
                "You should not need to fork a read-only session. Read-only sessions can be pickled and transmitted directly."
            )
        self._allow_changes = True
        return ForkSession(self._session)


class ForkSession(Session):
    def __getstate__(self) -> object:
        state = {"_session": self._session.as_bytes()}
        return state

    def __setstate__(self, state: object) -> None:
        if not isinstance(state, dict):
            raise ValueError("Invalid state")
        self._session = PySession.from_bytes(state["_session"])

    def merge(self, *others: Self) -> None:
        for other in others:
            if not isinstance(other, ForkSession):
                raise TypeError(
                    f"A ForkSession can only be merged with another ForkSession. Received {type(other)} instead."
                )
            self._session.merge(other._session)

    async def merge_async(self, *others: Self) -> None:
        """
        Merge the changes for this fork session with the changes from other fork sessions (async version).

        Parameters
        ----------
        others : ForkSession
            The other fork sessions to merge changes from.
        """
        for other in others:
            if not isinstance(other, ForkSession):
                raise TypeError(
                    f"A ForkSession can only be merged with another ForkSession. Received {type(other)} instead."
                )
            await self._session.merge_async(other._session)

    def commit(
        self,
        message: str,
        metadata: dict[str, Any] | None = None,
        rebase_with: ConflictSolver | None = None,
        rebase_tries: int = 1_000,
    ) -> NoReturn:
        raise TypeError(
            "Cannot commit a fork of a Session. If you are using uncooperative writes, "
            "please send the Repository object to your workers, not a Session. "
            "See https://icechunk.io/en/stable/icechunk-python/parallel/#distributed-writes for more."
        )

    async def commit_async(
        self,
        message: str,
        metadata: dict[str, Any] | None = None,
        rebase_with: ConflictSolver | None = None,
        rebase_tries: int = 1_000,
    ) -> NoReturn:
        raise TypeError(
            "Cannot commit a fork of a Session. If you are using uncooperative writes, "
            "please send the Repository object to your workers, not a Session. "
            "See https://icechunk.io/en/stable/icechunk-python/parallel/#distributed-writes for more."
        )

    @property
    def store(self) -> IcechunkStore:
        """
        Get a zarr Store object for reading and writing data from the repository using zarr python.

        Returns
        -------
        IcechunkStore
            A zarr Store object for reading and writing data from the repository.
        """
        return IcechunkStore(self._session.store, for_fork=True)



import importlib
from collections.abc import Hashable, Mapping, MutableMapping
from dataclasses import dataclass, field
from typing import Any, Literal, overload

import numpy as np
from packaging.version import Version

import xarray as xr
import zarr
from icechunk import IcechunkStore, Session
from icechunk.session import ForkSession
from icechunk.vendor.xarray import _choose_default_mode
from xarray import DataArray, Dataset
from xarray.backends.common import ArrayWriter
from xarray.backends.zarr import ZarrStore

__all__ = ["to_icechunk"]

Region = Mapping[str, slice | Literal["auto"]] | Literal["auto"] | None
ZarrWriteModes = Literal["w", "w-", "a", "a-", "r+", "r"]


try:
    has_dask = importlib.util.find_spec("dask") is not None
except ImportError:
    has_dask = False

if Version(xr.__version__) < Version("2024.10.0"):
    raise ValueError(
        f"Writing to icechunk requires Xarray>=2024.10.0 but you have {xr.__version__}. Please upgrade."
    )


def is_dask_collection(x: Any) -> bool:
    if has_dask:
        import dask

        return dask.base.is_dask_collection(x)
    else:
        return False


class LazyArrayWriter(ArrayWriter):
    def __init__(self) -> None:
        super().__init__()  # type: ignore[no-untyped-call]

        self.eager_sources: list[np.ndarray[Any, Any]] = []
        self.eager_targets: list[zarr.Array] = []
        self.eager_regions: list[tuple[slice, ...]] = []

    def add(self, source: Any, target: Any, region: Any = None) -> Any:
        if is_dask_collection(source):
            self.sources.append(source)
            self.targets.append(target)
            self.regions.append(region)
        else:
            self.eager_sources.append(source)
            self.eager_targets.append(target)
            self.eager_regions.append(region)

    def write_eager(self) -> None:
        for source, target, region in zip(
            self.eager_sources, self.eager_targets, self.eager_regions, strict=True
        ):
            target[region or ...] = source
        self.eager_sources = []
        self.eager_targets = []
        self.eager_regions = []


@dataclass
class _XarrayDatasetWriter:
    """
    Write Xarray Datasets to a group in an Icechunk store.

    This class is private API. Please do not use it.
    """

    dataset: Dataset = field(repr=False)
    store: IcechunkStore = field(kw_only=True)

    safe_chunks: bool = field(kw_only=True, default=True)

    _initialized: bool = field(default=False, repr=False)

    xarray_store: ZarrStore = field(init=False, repr=False)
    writer: LazyArrayWriter = field(init=False, repr=False)

    def __post_init__(self) -> None:
        if not isinstance(self.store, IcechunkStore):
            raise ValueError(
                f"Please pass in an icechunk.Session. Received {type(self.store)!r} instead."
            )

    def _open_group(
        self,
        *,
        group: str | None,
        mode: ZarrWriteModes | None,
        append_dim: Hashable | None,
        region: Region,
    ) -> None:
        concrete_mode: ZarrWriteModes = _choose_default_mode(
            mode=mode, append_dim=append_dim, region=region
        )

        self.xarray_store = ZarrStore.open_group(
            store=self.store,
            group=group,
            mode=concrete_mode,
            zarr_format=3,
            append_dim=append_dim,
            write_region=region,
            safe_chunks=self.safe_chunks,
            synchronizer=None,
            consolidated=False,
            consolidate_on_close=False,
            zarr_version=None,
        )
        self.dataset = self.xarray_store._validate_and_autodetect_region(self.dataset)

    def write_metadata(self, encoding: Mapping[Any, Any] | None = None) -> None:
        """
        This method creates new Zarr arrays when necessary, writes attributes,
        and any in-memory arrays.
        """
        from xarray.backends.api import _validate_dataset_names, dump_to_store

        # validate Dataset keys, DataArray names
        _validate_dataset_names(self.dataset)

        if encoding is None:
            encoding = {}
        self.xarray_store._validate_encoding(encoding)

        # This writes the metadata (zarr.json) for all arrays
        # This also will resize arrays for any appends
        self.writer = LazyArrayWriter()
        dump_to_store(self.dataset, self.xarray_store, self.writer, encoding=encoding)  # type: ignore[no-untyped-call]

        self._initialized = True

    def write_eager(self) -> None:
        """
        Write in-memory variables to store.

        Returns
        -------
        None
        """
        if not self._initialized:
            raise ValueError("Please call `write_metadata` first.")
        self.writer.write_eager()

    def write_lazy(
        self,
        chunkmanager_store_kwargs: MutableMapping[Any, Any] | None = None,
        split_every: int | None = None,
    ) -> ForkSession | None:
        """
        Write lazy arrays (e.g. dask) to store.
        """
        from icechunk.dask import session_merge_reduction

        if not self._initialized:
            raise ValueError("Please call `write_metadata` first.")

        if not self.writer.sources:
            return None

        chunkmanager_store_kwargs = chunkmanager_store_kwargs or {}
        chunkmanager_store_kwargs["load_stored"] = False
        chunkmanager_store_kwargs["return_stored"] = True

        # This calls dask.array.store, and we receive a dask array where each chunk is a Zarr array
        # each of those zarr.Array.store contains the changesets we need
        stored_arrays = self.writer.sync(
            compute=False, chunkmanager_store_kwargs=chunkmanager_store_kwargs
        )  # type: ignore[no-untyped-call]
        return session_merge_reduction(stored_arrays, split_every=split_every)


def to_icechunk(
    obj: DataArray | Dataset,
    session: Session,
    *,
    group: str | None = None,
    mode: ZarrWriteModes | None = None,
    safe_chunks: bool = True,
    append_dim: Hashable | None = None,
    region: Region = None,
    encoding: Mapping[Any, Any] | None = None,
    chunkmanager_store_kwargs: MutableMapping[Any, Any] | None = None,
    split_every: int | None = None,
) -> None:
    """
    Write an Xarray object to a group of an Icechunk store.

    Parameters
    ----------
    obj: DataArray or Dataset
        Xarray object to write
    session : icechunk.Session
        Writable Icechunk Session
    mode : {"w", "w-", "a", "a-", r+", None}, optional
        Persistence mode: "w" means create (overwrite if exists);
        "w-" means create (fail if exists);
        "a" means override all existing variables including dimension coordinates (create if does not exist);
        "a-" means only append those variables that have ``append_dim``.
        "r+" means modify existing array *values* only (raise an error if
        any metadata or shapes would change).
        The default mode is "a" if ``append_dim`` is set. Otherwise, it is
        "r+" if ``region`` is set and ``w-`` otherwise.
    group : str, optional
        Group path. (a.k.a. `path` in zarr terminology.)
    encoding : dict, optional
        Nested dictionary with variable names as keys and dictionaries of
        variable specific encodings as values, e.g.,
        ``{"my_variable": {"dtype": "int16", "scale_factor": 0.1,}, ...}``
    append_dim : hashable, optional
        If set, the dimension along which the data will be appended. All
        other dimensions on overridden variables must remain the same size.
    region : dict or "auto", optional
        Optional mapping from dimension names to either a) ``"auto"``, or b) integer
        slices, indicating the region of existing zarr array(s) in which to write
        this dataset's data.

        If ``"auto"`` is provided the existing store will be opened and the region
        inferred by matching indexes. ``"auto"`` can be used as a single string,
        which will automatically infer the region for all dimensions, or as
        dictionary values for specific dimensions mixed together with explicit
        slices for other dimensions.

        Alternatively integer slices can be provided; for example, ``{'x': slice(0,
        1000), 'y': slice(10000, 11000)}`` would indicate that values should be
        written to the region ``0:1000`` along ``x`` and ``10000:11000`` along
        ``y``.

        Users are expected to ensure that the specified region aligns with
        Zarr chunk boundaries, and that dask chunks are also aligned.
        Xarray makes limited checks that these multiple chunk boundaries line up.
        It is possible to write incomplete chunks and corrupt the data with this
        option if you are not careful.
    safe_chunks : bool, default: True
        If True, only allow writes to when there is a many-to-one relationship
        between Zarr chunks (specified in encoding) and Dask chunks.
        Set False to override this restriction; however, data may become corrupted
        if Zarr arrays are written in parallel.
        In addition to the many-to-one relationship validation, it also detects partial
        chunks writes when using the region parameter,
        these partial chunks are considered unsafe in the mode "r+" but safe in
        the mode "a".
        Note: Even with these validations it can still be unsafe to write
        two or more chunked arrays in the same location in parallel if they are
        not writing in independent regions.
    chunkmanager_store_kwargs : dict, optional
        Additional keyword arguments passed on to the `ChunkManager.store` method used to store
        chunked arrays. For example for a dask array additional kwargs will be passed eventually to
        `dask.array.store()`. Experimental API that should not be relied upon.
    split_every: int, optional
        Number of tasks to merge at every level of the tree reduction.

    Returns
    -------
    None

    Notes
    -----
    Two restrictions apply to the use of ``region``:

      - If ``region`` is set, _all_ variables in a dataset must have at
        least one dimension in common with the region. Other variables
        should be written in a separate single call to ``to_icechunk()``.
      - Dimensions cannot be included in both ``region`` and
        ``append_dim`` at the same time. To create empty arrays to fill
        in with ``region``, use the `_XarrayDatasetWriter` directly.
    """

    as_dataset = _make_dataset(obj)

    # This ugliness is needed so that we allow users to call `to_icechunk` with a dirty Session
    # for _serial_ writes
    is_dask = is_dask_collection(obj)
    fork: Session | ForkSession
    if is_dask:
        if session.has_uncommitted_changes:
            raise ValueError(
                "Calling `to_icechunk` is not allowed on a Session with uncommitted changes. Please commit first."
            )
        fork = session.fork()
    else:
        fork = session

    writer = _XarrayDatasetWriter(as_dataset, store=fork.store, safe_chunks=safe_chunks)

    writer._open_group(group=group, mode=mode, append_dim=append_dim, region=region)

    # write metadata
    writer.write_metadata(encoding)
    # write in-memory arrays
    writer.write_eager()
    # eagerly write dask arrays
    maybe_fork_session = writer.write_lazy(
        chunkmanager_store_kwargs=chunkmanager_store_kwargs
    )
    if is_dask:
        if maybe_fork_session is None:
            raise RuntimeError(
                "Logic bug! Please open at issue at https://github.com/earth-mover/icechunk"
            )
        session.merge(maybe_fork_session)
    else:
        if maybe_fork_session is not None:
            raise RuntimeError(
                "Unexpected write of dask arrays! Please open at issue at https://github.com/earth-mover/icechunk"
            )


@overload
def _make_dataset(obj: DataArray) -> Dataset: ...
@overload
def _make_dataset(obj: Dataset) -> Dataset: ...
def _make_dataset(obj: DataArray | Dataset) -> Dataset:
    """Copied from DataArray.to_zarr"""
    DATAARRAY_NAME = "__xarray_dataarray_name__"
    DATAARRAY_VARIABLE = "__xarray_dataarray_variable__"

    if isinstance(obj, Dataset):
        return obj

    assert isinstance(obj, DataArray)

    if obj.name is None:
        # If no name is set then use a generic xarray name
        dataset = obj.to_dataset(name=DATAARRAY_VARIABLE)
    elif obj.name in obj.coords or obj.name in obj.dims:
        # The name is the same as one of the coords names, which the netCDF data model
        # does not support, so rename it but keep track of the old name
        dataset = obj.to_dataset(name=DATAARRAY_VARIABLE)
        dataset.attrs[DATAARRAY_NAME] = obj.name
    else:
        # No problems with the name - so we're fine!
        dataset = obj.to_dataset()
    return dataset



from icechunk import IcechunkStore
from tests.conftest import parse_repo
from zarr.storage._common import make_store_path


async def test_make_store_path() -> None:
    # Memory store
    repo = parse_repo("memory", path="")
    session = repo.writable_session("main")
    store = session.store
    store_path = await make_store_path(store)
    assert isinstance(store_path.store, IcechunkStore)



import pytest

pytest.importorskip("dask")
pytest.importorskip("distributed")

import tempfile

import dask
import distributed
import xarray as xr
import zarr
from icechunk import Repository, local_filesystem_storage
from icechunk.dask import store_dask
from icechunk.xarray import to_icechunk
from tests.test_xarray import create_test_data, roundtrip
from xarray.testing import assert_identical


def test_store_dask() -> None:
    shape = (100, 100)
    dask_chunks = (20, 20)
    dask_array = dask.array.random.random(shape, chunks=dask_chunks)

    zarr_chunks = (10, 10)
    with tempfile.TemporaryDirectory() as tmpdir:
        repo = Repository.create(local_filesystem_storage(tmpdir))
        session = repo.writable_session("main")
        group = zarr.group(store=session.store, overwrite=True)

        zarray = group.create_array(
            "array",
            shape=shape,
            chunks=zarr_chunks,
            dtype="f8",
            fill_value=float("nan"),
        )
        # can't fork a dirty session
        with dask.config.set(scheduler="processes"):
            with pytest.raises(ValueError):
                store_dask(sources=[dask_array], targets=[zarray])

        # session = repo.writable_session("main")
        # Can't fork a dirty session
        with pytest.raises(ValueError):
            fork = session.fork()

        session.commit("i am forced to commit to make progress")
        session = repo.writable_session("main")
        fork = session.fork()
        zarray = zarr.open_array(fork.store, path="array")
        with dask.config.set(scheduler="processes"):
            store_dask(sources=[dask_array], targets=[zarray])


def test_distributed() -> None:
    with distributed.Client():  # type: ignore [no-untyped-call]
        ds = create_test_data().chunk(dim1=3, dim2=4)
        with roundtrip(ds, commit=True) as actual:
            assert_identical(actual, ds)


@pytest.mark.parametrize("scheduler", ["threads", "processes"])
def test_dask_schedulers(scheduler) -> None:
    with dask.config.set(scheduler=scheduler):
        ds = create_test_data().chunk(dim1=3, dim2=4)
        with roundtrip(ds, commit=scheduler == "processes") as actual:
            assert_identical(actual, ds)


@pytest.mark.parametrize("scheduler", ["threads", "processes"])
def test_xarray_to_icechunk_nested_pickling(scheduler) -> None:
    with dask.config.set(scheduler=scheduler):
        ds = create_test_data(dim_sizes=(2, 3, 4)).chunk(-1)
        with tempfile.TemporaryDirectory() as tmpdir:
            repo = Repository.create(local_filesystem_storage(tmpdir))
            session = repo.writable_session("main")

            to_icechunk(ds, session=session, mode="w")
            with pytest.raises(ValueError, match="Please commit first"):
                to_icechunk(ds, session=session, mode="w")
            # Needed because we can't pickle the writable session for distributed read
            session.commit("wrote a commit.")
            with xr.open_zarr(session.store, consolidated=False) as actual:
                assert_identical(actual, ds)

            newds = ds + 1
            session = repo.writable_session("main")
            to_icechunk(newds, session=session, mode="w")
            # Needed because we can't pickle the writable session
            session.commit("wrote another commit.")
            with xr.open_zarr(session.store, consolidated=False) as actual:
                assert_identical(actual, newds)



import operator

import pytest

import icechunk as ic
import xarray as xr
import zarr
from benchmarks.datasets import Dataset

# TODO: configurable?
zarr.config.set({"async.concurrency": 64})

pytestmark = pytest.mark.read_benchmark


@pytest.fixture(
    params=[
        pytest.param(None, id="default"),
        pytest.param(
            ic.ManifestPreloadConfig(preload_if=ic.ManifestPreloadCondition.false()),
            id="off",
        ),
    ],
)
def preload(request):
    return request.param


@pytest.mark.setup_benchmarks
@pytest.mark.filterwarnings("ignore:datetime.datetime.utcnow")  # bah, boto!
def test_recreate_datasets(synth_dataset, request):
    """
    Helper to (re)create Icechunk repos when format or Dataset changes.
    Use `pytest -m setup_benchmarks` to run it
    `--force-setup=True` will always recreate.
    `--force-setup=False` will recreate if icechunk cannot create a Repository (perhaps due to format change).
    """
    # Check if the mark is passed via the command line
    if "setup_benchmarks" not in request.config.getoption("-m", default=""):
        pytest.skip(
            "Skipping re-creating benchmarks because 'setup_benchmarks' is not set in the command line."
        )

    if synth_dataset.setupfn is not None:
        force = request.config.getoption("--force-setup") == "True"
        synth_dataset.setup(force=force)


def test_time_create_store(synth_dataset: Dataset, benchmark) -> None:
    """time to create the icechunk create the Repo, session, and store objects."""
    benchmark(operator.attrgetter("store"), synth_dataset)


@pytest.mark.benchmark(warmup=1)
def test_time_getsize_key(synth_dataset: Dataset, benchmark) -> None:
    from zarr.core.sync import sync

    if synth_dataset.load_variables is None:
        pytest.skip()

    store = synth_dataset.store

    @benchmark
    def fn():
        for array in synth_dataset.load_variables:
            if (group := synth_dataset.group) is not None:
                prefix = f"{group}/"
            else:
                prefix = ""
            key = f"{prefix}{array}/zarr.json"
            sync(store.getsize(key))


def test_time_getsize_prefix(synth_dataset: Dataset, benchmark) -> None:
    if synth_dataset.first_byte_variable is None:
        pytest.skip()
    store = synth_dataset.store
    group = zarr.open_group(store, path=synth_dataset.group, mode="r")
    # Using a relatively small variable.
    array = group[synth_dataset.first_byte_variable]
    benchmark(operator.methodcaller("nbytes_stored"), array)


@pytest.mark.benchmark(group="zarr-read")
def test_time_zarr_open(synth_dataset: Dataset, benchmark) -> None:
    """
    We repeatedly create the store inside the benchmarked function to
    test the "cold open" time, and downloading of the snapshot file.
    """

    @benchmark
    def fn():
        zarr.open_group(synth_dataset.store, path=synth_dataset.group, mode="r")


@pytest.mark.benchmark(group="zarr-read")
def test_time_zarr_members(synth_dataset: Dataset, benchmark) -> None:
    # list_dir, maybe warmup=1
    group = zarr.open_group(synth_dataset.store, path=synth_dataset.group, mode="r")
    benchmark(operator.methodcaller("members"), group)


@pytest.mark.benchmark(group="xarray-read", min_rounds=10)
def test_time_xarray_open(synth_dataset: Dataset, benchmark) -> None:
    @benchmark
    def fn():
        xr.open_zarr(
            synth_dataset.store,
            group=synth_dataset.group,
            chunks=None,
            consolidated=False,
        )


@pytest.mark.parametrize("selector", ["single-chunk", "full-read"])
@pytest.mark.benchmark(group="xarray-read", min_rounds=2)
def test_time_xarray_read_chunks_cold_cache(
    synth_dataset: Dataset, selector, preload, benchmark
) -> None:
    """128MB vs 8MB chunks. should see a difference."""
    if synth_dataset.load_variables is None:
        pytest.skip()

    if selector == "single-chunk":
        subsetter = synth_dataset.chunk_selector
    elif selector == "full-read":
        subsetter = synth_dataset.full_load_selector
    else:
        raise ValueError(f"bad selector: {selector}")

    if subsetter is None:
        pytest.skip(f"no selector specified for {selector}")

    # TODO: switch out concurrency "ideal_request_size"
    @benchmark
    def fn():
        repo = ic.Repository.open(
            storage=synth_dataset.storage,
            config=ic.RepositoryConfig(manifest=ic.ManifestConfig(preload=preload)),
        )
        ds = xr.open_zarr(
            repo.readonly_session("main").store,
            group=synth_dataset.group,
            chunks=None,
            consolidated=False,
        )
        subset = ds.isel(subsetter)
        subset[synth_dataset.load_variables].compute()


@pytest.mark.benchmark(group="xarray-read", min_rounds=2)
def test_time_xarray_read_chunks_hot_cache(
    synth_dataset: Dataset, preload, benchmark
) -> None:
    """128MB vs 8MB chunks. should see a difference."""
    if synth_dataset.load_variables is None:
        pytest.skip()
    # TODO: switch out concurrency "ideal_request_size"
    repo = ic.Repository.open(
        storage=synth_dataset.storage,
        config=ic.RepositoryConfig(manifest=ic.ManifestConfig(preload=preload)),
    )
    ds = xr.open_zarr(
        repo.readonly_session("main").store,
        group=synth_dataset.group,
        chunks=None,
        consolidated=False,
    )
    subset = ds.isel(synth_dataset.chunk_selector)
    # important this cannot be `load`
    benchmark(operator.methodcaller("compute"), subset[synth_dataset.load_variables])


@pytest.mark.benchmark(group="bytes-read")
def test_time_first_bytes(synth_dataset: Dataset, benchmark) -> None:
    """TODO: this should be sensitive to manifest splitting"""
    if synth_dataset.first_byte_variable is None:
        pytest.skip("first_byte_variable not set!")

    @benchmark
    def open_and_read():
        # by opening the group repeatedly inside the benchmarked function
        # we force re-download of manifest so that we actually measure what we want.
        group = zarr.open_group(synth_dataset.store, path=synth_dataset.group, mode="r")
        group[synth_dataset.first_byte_variable][:]


# TODO: synthetic dataset with very deep and large hierarchies for e.g. tree & members



from collections.abc import Generator

import pytest

pytest.importorskip("xarray")

import contextlib
import string
import tempfile

import numpy as np
import pandas as pd

import xarray as xr
from icechunk import Repository, in_memory_storage, local_filesystem_storage
from icechunk.xarray import to_icechunk
from xarray.testing import assert_identical


def create_test_data(
    seed: int | None = None,
    add_attrs: bool = True,
    dim_sizes: tuple[int, int, int] = (8, 9, 10),
) -> xr.Dataset:
    rs = np.random.RandomState(seed)
    _vars = {
        "var1": ["dim1", "dim2"],
        "var2": ["dim1", "dim2"],
        "var3": ["dim3", "dim1"],
    }
    _dims = {"dim1": dim_sizes[0], "dim2": dim_sizes[1], "dim3": dim_sizes[2]}

    obj = xr.Dataset()
    obj["dim2"] = ("dim2", 0.5 * np.arange(_dims["dim2"]))
    if _dims["dim3"] > 26:
        raise RuntimeError(
            f"Not enough letters for filling this dimension size ({_dims['dim3']})"
        )
    obj["dim3"] = ("dim3", list(string.ascii_lowercase[0 : _dims["dim3"]]))
    obj["time"] = ("time", pd.date_range("2000-01-01", periods=20))
    for v, dims in sorted(_vars.items()):
        data = rs.normal(size=tuple(_dims[d] for d in dims))
        obj[v] = (dims, data)
        if add_attrs:
            obj[v].attrs = {"foo": "variable"}
    numbers_values = rs.randint(0, 3, _dims["dim3"], dtype="int64")
    obj.coords["numbers"] = ("dim3", numbers_values)
    obj.encoding = {"foo": "bar"}
    return obj


@contextlib.contextmanager
def roundtrip(
    data: xr.Dataset, *, commit: bool = False
) -> Generator[xr.Dataset, None, None]:
    with tempfile.TemporaryDirectory() as tmpdir:
        repo = Repository.create(local_filesystem_storage(tmpdir))
        session = repo.writable_session("main")
        to_icechunk(data, session=session, mode="w")
        session.commit("write")
        with xr.open_zarr(session.store, consolidated=False) as ds:
            yield ds


def test_xarray_to_icechunk() -> None:
    ds = create_test_data()
    with roundtrip(ds) as actual:
        assert_identical(actual, ds)


def test_repeated_to_icechunk_serial() -> None:
    ds = create_test_data()
    repo = Repository.create(in_memory_storage())
    session = repo.writable_session("main")
    to_icechunk(ds, session)
    to_icechunk(ds, session, mode="w")



import asyncio
import concurrent.futures
import random
import time
import uuid
from random import randrange
from threading import Event

import pytest
from termcolor import colored

import icechunk
import zarr
from tests.conftest import write_chunks_to_minio

N = 15


async def write_to_store(
    array: zarr.Array, x: int, y: int, barrier: asyncio.Barrier
) -> None:
    await barrier.wait()
    await asyncio.sleep(random.uniform(0, 0.5))
    array[x, y] = x * y
    # await asyncio.sleep(0)


async def read_store(array: zarr.Array, x: int, y: int, barrier: asyncio.Barrier) -> None:
    await barrier.wait()
    while True:
        value = array[x, y]
        if value == x * y:
            break
        await asyncio.sleep(random.uniform(0, 0.1))


async def list_store(store: icechunk.IcechunkStore, barrier: asyncio.Barrier) -> None:
    expected = set(
        ["zarr.json", "array/zarr.json"]
        + [f"array/c/{x}/{y}" for x in range(N) for y in range(N - 1)]
    )
    await barrier.wait()
    while True:
        current: set[str] | None = set([k async for k in store.list_prefix("")])
        if current == expected:
            break
        current = None
        await asyncio.sleep(0.1)


async def test_concurrency() -> None:
    repo = icechunk.Repository.open_or_create(
        storage=icechunk.in_memory_storage(),
    )

    session = repo.writable_session("main")
    store = session.store

    group = zarr.group(store=store, overwrite=True)
    array = group.create_array(
        "array", shape=(N, N), chunks=(1, 1), dtype="f8", fill_value=1e23
    )

    barrier = asyncio.Barrier(2 * N * (N - 1) + 1)

    async with asyncio.TaskGroup() as tg:
        _task1 = tg.create_task(list_store(store, barrier), name="listing")

        for x in range(N):
            for y in range(N - 1):
                _write_task = tg.create_task(
                    read_store(array, x, y, barrier), name=f"read {x},{y}"
                )

        for x in range(N):
            for y in range(N - 1):
                _write_task = tg.create_task(
                    write_to_store(array, x, y, barrier), name=f"write {x},{y}"
                )

    all_coords = {coords async for coords in session.chunk_coordinates("/array")}
    assert all_coords == {(x, y) for x in range(N) for y in range(N - 1)}

    _res = session.commit("commit")

    assert isinstance(group["array"], zarr.Array)
    array = group["array"]
    assert isinstance(array, zarr.Array)

    for x in range(N):
        for y in range(N - 1):
            assert array[x, y] == x * y


@pytest.mark.filterwarnings("ignore:datetime.datetime.utcnow")
async def test_thread_concurrency() -> None:
    """Run multiple threads doing different type of operations for SECONDS_TO_RUN seconds.

    The threads execute 5 types of operations: reads, native writes, virtual writes, deletes and lists.

    We launch THREADS threads of each type, and at the end we assert all operation types executed
    a few times.

    The output prints a block character for each operation, colored according to the operation type.
    The expectation is blocks of different colors should interleave.
    """
    THREADS = 20
    SECONDS_TO_RUN = 1

    prefix = str(uuid.uuid4())
    write_chunks_to_minio(
        [
            (f"{prefix}/chunk-1", b"first"),
            (f"{prefix}/chunk-2", b"second"),
        ],
    )

    config = icechunk.RepositoryConfig.default()
    store_config = icechunk.s3_store(
        region="us-east-1",
        endpoint_url="http://localhost:9000",
        allow_http=True,
        force_path_style=True,
        s3_compatible=True,
    )
    container = icechunk.VirtualChunkContainer("s3://testbucket/", store_config)
    config.set_virtual_chunk_container(container)
    config.inline_chunk_threshold_bytes = 0
    credentials = icechunk.containers_credentials(
        {
            "s3://testbucket": icechunk.s3_credentials(
                access_key_id="minio123", secret_access_key="minio123"
            )
        }
    )

    storage = icechunk.s3_storage(
        region="us-east-1",
        endpoint_url="http://localhost:9000",
        allow_http=True,
        force_path_style=True,
        bucket="testbucket",
        prefix="multithreaded-test__" + str(time.time()),
        access_key_id="minio123",
        secret_access_key="minio123",
    )

    # Open the store
    repo = icechunk.Repository.create(
        storage=storage,
        config=config,
        authorize_virtual_chunk_access=credentials,
    )

    session = repo.writable_session("main")
    store = session.store

    group = zarr.group(store=store, overwrite=True)
    group.create_array("array", shape=(1_000,), chunks=(1,), dtype="i4", compressors=None)

    def do_virtual_writes(start, stop) -> int:
        n = 0
        start.wait()
        while not stop.is_set():
            i = randrange(1_000)
            store.set_virtual_ref(
                f"array/c/{i}",
                f"s3://testbucket/{prefix}/chunk-1",
                offset=0,
                length=4,
            )
            print(colored("■", "green"), end="")
            n += 1
        return n

    def do_native_writes(start, stop) -> int:
        async def do() -> int:
            n = 0
            while not stop.is_set():
                i = randrange(1_000)
                await store.set(
                    f"array/c/{i}", zarr.core.buffer.cpu.Buffer.from_bytes(b"0123")
                )
                print(colored("■", "white"), end="")
                n += 1
            return n

        start.wait()
        return asyncio.run(do())

    def do_reads(start, stop) -> int:
        buffer_prototype = zarr.core.buffer.default_buffer_prototype()

        async def do() -> int:
            n = 0
            while not stop.is_set():
                i = randrange(1_000)
                await store.get(f"array/c/{i}", prototype=buffer_prototype)
                print(colored("■", "blue"), end="")
                n += 1
            return n

        start.wait()
        return asyncio.run(do())

    def do_deletes(start, stop) -> int:
        async def do() -> int:
            n = 0
            while not stop.is_set():
                i = randrange(1_000)
                await store.delete(f"array/c/{i}")
                print(colored("■", "red"), end="")
                n += 1
            return n

        start.wait()
        return asyncio.run(do())

    def do_lists(start, stop) -> int:
        async def do() -> int:
            n = 0
            while not stop.is_set():
                _ = [k async for k in store.list_prefix("")]
                print(colored("■", "yellow"), end="")
                n += 1
            return n

        start.wait()
        return asyncio.run(do())

    with concurrent.futures.ThreadPoolExecutor(max_workers=THREADS * 5) as pool:
        virtual_writes = []
        native_writes = []
        deletes = []
        reads = []
        lists = []

        start = Event()
        stop = Event()

        for _ in range(THREADS):
            virtual_writes.append(pool.submit(do_virtual_writes, start, stop))
            native_writes.append(pool.submit(do_native_writes, start, stop))
            deletes.append(pool.submit(do_deletes, start, stop))
            reads.append(pool.submit(do_reads, start, stop))
            lists.append(pool.submit(do_lists, start, stop))

        start.set()
        time.sleep(SECONDS_TO_RUN)
        stop.set()

        virtual_writes_count = sum(future.result() for future in virtual_writes)
        native_writes_count = sum(future.result() for future in native_writes)
        deletes_count = sum(future.result() for future in deletes)
        reads_count = sum(future.result() for future in reads)
        lists_count = sum(future.result() for future in lists)

    print()
    print(
        f"virtual writes: {virtual_writes_count}, native writes: {native_writes_count}, deletes: {deletes_count}, reads: {reads_count}, lists: {lists_count}"
    )

    assert virtual_writes_count > 2
    assert native_writes_count > 2
    assert deletes_count > 2
    assert reads_count > 2
    assert lists_count > 2



# Deepak runs this with
# pip install ../../xarray && ICECHUNK_XARRAY_BACKENDS_TESTS=1 pytest -s --pdb -c=../../xarray/pyproject.toml -W ignore tests/run_xarray_backends_tests.py
import contextlib
import os
import pickle
import tempfile
import time
from collections.abc import Generator
from typing import Any

import numpy as np
import pytest

import xarray as xr
import zarr
from icechunk import (
    IcechunkStore,
    Repository,
    in_memory_storage,
    local_filesystem_storage,
    s3_storage,
)
from icechunk.xarray import to_icechunk
from xarray.testing import assert_identical

# needed otherwise not discovered
from xarray.tests.conftest import time_unit  # noqa: F401, RUF100
from xarray.tests.test_backends import (
    TestZarrRegionAuto as ZarrRegionAutoTests,
)
from xarray.tests.test_backends import (
    ZarrBase,
    # needed otherwise not discovered
    default_zarr_format,  # noqa: F401, RUF100
)


@pytest.mark.skipif(
    not os.environ.get("ICECHUNK_XARRAY_BACKENDS_TESTS", None),
    reason="skipping Xarray backends tests",
)
class IcechunkStoreBase(ZarrBase):
    @contextlib.contextmanager
    def create_repo(self) -> Generator[Repository]:
        raise NotImplementedError

    @contextlib.contextmanager
    def create_zarr_target(self) -> Generator[IcechunkStore]:
        if zarr.config.config["default_zarr_format"] == 2:
            pytest.skip("v2 not supported")
        with self.create_repo() as repo:
            session = repo.writable_session("main")
            yield session.store

    @pytest.mark.parametrize("consolidated", [False, True, None])
    @pytest.mark.parametrize("compute", [False, True])
    @pytest.mark.parametrize("use_dask", [False, True])
    @pytest.mark.parametrize("write_empty", [False, True, None])
    def test_write_region(
        self, consolidated: Any, compute: Any, use_dask: Any, write_empty: Any
    ) -> None:
        if consolidated is not False:
            pytest.skip("consolidated not supported.")
        super().test_write_region(consolidated, compute, use_dask, write_empty)

    @pytest.mark.parametrize("consolidated", [False, True, None])
    def test_roundtrip_consolidated(self, consolidated: Any) -> None:
        if consolidated is not False:
            pytest.skip("consolidated not supported.")
        super().test_roundtrip_consolidated(consolidated)

    @contextlib.contextmanager
    def roundtrip_with_commit(self, ds: xr.Dataset) -> Generator[xr.Dataset]:
        with self.create_repo() as repo:
            session = repo.writable_session("main")
            to_icechunk(ds, session)
            session.commit("initial commit")
            yield xr.open_zarr(session.store)

    def test_pickle(self) -> None:
        expected = xr.Dataset({"foo": ("x", [42])})
        with self.roundtrip_with_commit(expected) as roundtripped:
            raw_pickle = pickle.dumps(roundtripped)
            with pickle.loads(raw_pickle) as unpickled_ds:
                assert_identical(expected, unpickled_ds)

    @pytest.mark.filterwarnings("ignore:deallocating CachingFileManager")
    def test_pickle_dataarray(self) -> None:
        expected = xr.Dataset({"foo": ("x", [42])})
        with self.roundtrip_with_commit(expected) as roundtripped:
            raw_pickle = pickle.dumps(roundtripped["foo"])
            # TODO: figure out how to explicitly close the file for the
            # unpickled DataArray?
            unpickled = pickle.loads(raw_pickle)
            assert_identical(expected["foo"], unpickled)


class TestIcechunkStoreFilesystem(IcechunkStoreBase):
    @contextlib.contextmanager
    def create_repo(self) -> Generator[Repository]:
        with tempfile.TemporaryDirectory() as tmpdir:
            yield Repository.create(local_filesystem_storage(tmpdir))


class TestIcechunkStoreMemory(IcechunkStoreBase):
    @contextlib.contextmanager
    def create_repo(self) -> Generator[Repository]:
        yield Repository.create(in_memory_storage())

    def test_pickle(self) -> None:
        pytest.skip("pickling memory store is not supported")

    def test_pickle_dataarray(self) -> None:
        pytest.skip("pickling memory store is not supported")


class TestIcechunkStoreMinio(IcechunkStoreBase):
    @contextlib.contextmanager
    def create_repo(self) -> Generator[Repository]:
        yield Repository.create(
            s3_storage(
                endpoint_url="http://localhost:9000",
                allow_http=True,
                force_path_style=True,
                region="us-east-1",
                bucket="testbucket",
                prefix="python-xarray-test__" + str(time.time()),
                access_key_id="minio123",
                secret_access_key="minio123",
            )
        )


@pytest.mark.filterwarnings("ignore:Failed to open:RuntimeWarning")
class TestIcechunkRegionAuto(ZarrRegionAutoTests):
    @contextlib.contextmanager
    def create_zarr_target(self) -> Generator[IcechunkStore]:
        if zarr.config.config["default_zarr_format"] == 2:
            pytest.skip("v2 not supported")
        repo = Repository.create(in_memory_storage())
        session = repo.writable_session("main")
        yield session.store

    @contextlib.contextmanager
    def create(self):
        x = np.arange(0, 50, 10)
        y = np.arange(0, 20, 2)
        data = np.ones((5, 10))
        ds = xr.Dataset(
            {"test": xr.DataArray(data, dims=("x", "y"), coords={"x": x, "y": y})}
        )
        repo = Repository.create(in_memory_storage())
        session = repo.writable_session("main")
        self.save(session.store, ds)
        session.commit("initial commit")
        yield repo.writable_session("main").store, ds

    def save(self, target, ds, **kwargs):
        # not really important here
        kwargs.pop("compute", None)
        to_icechunk(ds, session=target.session, **kwargs)

    def test_zarr_append_chunk_partial(self):
        pytest.skip(
            "this test requires multiple saves, and is meant to exercise Xarray logic."
        )

    def test_zarr_safe_chunk_region(self, *args, **kwargs):
        pytest.skip(
            "this test requires multiple saves, and is meant to exercise Xarray logic."
        )

    def test_zarr_safe_chunk_append_dim(self, *args, **kwargs):
        pytest.skip(
            "this test requires multiple saves, and is meant to exercise Xarray logic."
        )

    def test_zarr_region_chunk_partial_offset(self):
        pytest.skip(
            "this test requires multiple saves, and is meant to exercise Xarray logic."
        )



from collections.abc import AsyncIterator, Iterable
from datetime import datetime
from typing import TYPE_CHECKING, Any

from icechunk._icechunk_python import PyStore, VirtualChunkSpec
from zarr.abc.store import (
    ByteRequest,
    OffsetByteRequest,
    RangeByteRequest,
    Store,
    SuffixByteRequest,
)
from zarr.core.buffer import Buffer, BufferPrototype
from zarr.core.common import BytesLike
from zarr.core.sync import SyncMixin

if TYPE_CHECKING:
    from icechunk import Session


def _byte_request_to_tuple(
    byte_request: ByteRequest | None,
) -> tuple[int | None, int | None]:
    match byte_request:
        case None:
            return (None, None)
        case RangeByteRequest(start, end):
            return (start, end)
        case OffsetByteRequest(offset):
            return (offset, None)
        case SuffixByteRequest(suffix):
            return (None, suffix)
        case _:
            raise ValueError(f"Unexpected byte_range, got {byte_request}")


class IcechunkStore(Store, SyncMixin):
    _store: PyStore
    _for_fork: bool

    def __init__(
        self,
        store: PyStore,
        for_fork: bool,
        read_only: bool | None = None,
        *args: Any,
        **kwargs: Any,
    ):
        """Create a new IcechunkStore.

        This should not be called directly, instead use the `create`, `open_existing` or `open_or_create` class methods.
        """
        read_only = read_only if read_only is not None else store.read_only
        super().__init__(read_only=read_only)
        if store is None:
            raise ValueError(
                "An IcechunkStore should not be created with the default constructor, instead use either the create or open_existing class methods."
            )
        self._store = store
        self._is_open = True
        self._for_fork = for_fork

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, IcechunkStore):
            return False
        return self._store == value._store

    def __getstate__(self) -> object:
        # for read_only sessions we allow pickling, this allows distributed reads without forking
        writable = not self.session.read_only
        if writable and not self._for_fork:
            raise ValueError(
                "You must opt-in to pickle writable sessions in a distributed context "
                "using Session.fork(). "
                # link to docs
                "If you are using xarray's `Dataset.to_zarr` method to write dask arrays, "
                "please use `icechunk.xarray.to_icechunk` instead. "
            )
        d = self.__dict__.copy()
        # we serialize the Rust store as bytes
        d["_store"] = self._store.as_bytes()
        d["_for_fork"] = self._for_fork
        return d

    def __setstate__(self, state: Any) -> None:
        # we have to deserialize the bytes of the Rust store
        store_repr = state["_store"]
        state["_store"] = PyStore.from_bytes(store_repr)
        self.__dict__ = state

    def with_read_only(self, read_only: bool = False) -> Store:
        new_store = IcechunkStore(store=self._store, for_fork=False, read_only=read_only)
        new_store._is_open = False
        return new_store

    @property
    def session(self) -> "Session":
        from icechunk.session import ForkSession, Session

        if self._for_fork:
            return ForkSession(self._store.session)
        else:
            return Session(self._store.session)

    async def clear(self) -> None:
        """Clear the store.

        This will remove all contents from the current session,
        including all groups and all arrays. But it will not modify the repository history.
        """
        return await self._store.clear()

    def sync_clear(self) -> None:
        """Clear the store.

        This will remove all contents from the current session,
        including all groups and all arrays. But it will not modify the repository history.
        """
        return self._store.sync_clear()

    async def is_empty(self, prefix: str) -> bool:
        """
        Check if the directory is empty.

        Parameters
        ----------
        prefix : str
            Prefix of keys to check.

        Returns
        -------
        bool
            True if the store is empty, False otherwise.
        """
        return await self._store.is_empty(prefix)

    async def get(
        self,
        key: str,
        prototype: BufferPrototype,
        byte_range: ByteRequest | None = None,
    ) -> Buffer | None:
        """Retrieve the value associated with a given key.

        Parameters
        ----------
        key : str
        byte_range : ByteRequest, optional

            ByteRequest may be one of the following. If not provided, all data associated with the key is retrieved.

            - RangeByteRequest(int, int): Request a specific range of bytes in the form (start, end). The end is exclusive. If the given range is zero-length or starts after the end of the object, an error will be returned. Additionally, if the range ends after the end of the object, the entire remainder of the object will be returned. Otherwise, the exact requested range will be returned.
            - OffsetByteRequest(int): Request all bytes starting from a given byte offset. This is equivalent to bytes={int}- as an HTTP header.
            - SuffixByteRequest(int): Request the last int bytes. Note that here, int is the size of the request, not the byte offset. This is equivalent to bytes=-{int} as an HTTP header.

        Returns
        -------
        Buffer
        """

        try:
            result = await self._store.get(key, _byte_request_to_tuple(byte_range))
        except KeyError as _e:
            # Zarr python expects None to be returned if the key does not exist
            # but an IcechunkStore returns an error if the key does not exist
            return None

        return prototype.buffer.from_bytes(result)

    async def get_partial_values(
        self,
        prototype: BufferPrototype,
        key_ranges: Iterable[tuple[str, ByteRequest | None]],
    ) -> list[Buffer | None]:
        """Retrieve possibly partial values from given key_ranges.

        Parameters
        ----------
        key_ranges : Iterable[tuple[str, tuple[int | None, int | None]]]
            Ordered set of key, range pairs, a key may occur multiple times with different ranges

        Returns
        -------
        list of values, in the order of the key_ranges, may contain null/none for missing keys
        """
        # NOTE: pyo3 has not implicit conversion from an Iterable to a rust iterable. So we convert it
        # to a list here first. Possible opportunity for optimization.
        ranges = [(k[0], _byte_request_to_tuple(k[1])) for k in key_ranges]
        result = await self._store.get_partial_values(list(ranges))
        return [prototype.buffer.from_bytes(r) for r in result]

    async def exists(self, key: str) -> bool:
        """Check if a key exists in the store.

        Parameters
        ----------
        key : str

        Returns
        -------
        bool
        """
        return await self._store.exists(key)

    @property
    def supports_writes(self) -> bool:
        """Does the store support writes?"""
        return self._store.supports_writes

    async def set(self, key: str, value: Buffer) -> None:
        """Store a (key, value) pair.

        Parameters
        ----------
        key : str
        value : Buffer
        """
        if not isinstance(value, Buffer):
            raise TypeError(
                f"IcechunkStore.set(): `value` must be a Buffer instance. Got an instance of {type(value)} instead."
            )
        return await self._store.set(key, value.to_bytes())

    async def set_if_not_exists(self, key: str, value: Buffer) -> None:
        """
        Store a key to ``value`` if the key is not already present.

        Parameters
        -----------
        key : str
        value : Buffer
        """
        return await self._store.set_if_not_exists(key, value.to_bytes())

    def set_virtual_ref(
        self,
        key: str,
        location: str,
        *,
        offset: int,
        length: int,
        checksum: str | datetime | None = None,
        validate_container: bool = True,
    ) -> None:
        """Store a virtual reference to a chunk.

        Parameters
        ----------
        key : str
            The chunk to store the reference under. This is the fully qualified zarr key eg: 'array/c/0/0/0'
        location : str
            The location of the chunk in storage. This is absolute path to the chunk in storage eg: 's3://bucket/path/to/file.nc'
        offset : int
            The offset in bytes from the start of the file location in storage the chunk starts at
        length : int
            The length of the chunk in bytes, measured from the given offset
        checksum : str | datetime | None
            The etag or last_medified_at field of the object
        validate_container: bool
            If set to true, fail for locations that don't match any existing virtual chunk container
        """
        return self._store.set_virtual_ref(
            key, location, offset, length, checksum, validate_container
        )

    async def set_virtual_ref_async(
        self,
        key: str,
        location: str,
        *,
        offset: int,
        length: int,
        checksum: str | datetime | None = None,
        validate_container: bool = True,
    ) -> None:
        """Store a virtual reference to a chunk asynchronously.

        Parameters
        ----------
        key : str
            The chunk to store the reference under. This is the fully qualified zarr key eg: 'array/c/0/0/0'
        location : str
            The location of the chunk in storage. This is absolute path to the chunk in storage eg: 's3://bucket/path/to/file.nc'
        offset : int
            The offset in bytes from the start of the file location in storage the chunk starts at
        length : int
            The length of the chunk in bytes, measured from the given offset
        checksum : str | datetime | None
            The etag or last_medified_at field of the object
        validate_container: bool
            If set to true, fail for locations that don't match any existing virtual chunk container
        """
        return await self._store.set_virtual_ref_async(
            key, location, offset, length, checksum, validate_container
        )

    def set_virtual_refs(
        self,
        array_path: str,
        chunks: list[VirtualChunkSpec],
        *,
        validate_containers: bool = True,
    ) -> list[tuple[int, ...]] | None:
        """Store multiple virtual references for the same array.

        Parameters
        ----------
        array_path : str
            The path to the array inside the Zarr store. Example: "/groupA/groupB/outputs/my-array"
        chunks : list[VirtualChunkSpec],
            The list of virtual chunks to add
        validate_containers: bool
            If set to true, ignore virtual references for locations that don't match any existing virtual chunk container


        Returns
        -------
        list[tuple[int, ...]] | None

            If all virtual references where successfully updated, it returns None.
            If there were validation errors, it returns the chunk indices of all failed references.
        """
        return self._store.set_virtual_refs(array_path, chunks, validate_containers)

    async def set_virtual_refs_async(
        self,
        array_path: str,
        chunks: list[VirtualChunkSpec],
        *,
        validate_containers: bool = True,
    ) -> list[tuple[int, ...]] | None:
        """Store multiple virtual references for the same array asynchronously.

        Parameters
        ----------
        array_path : str
            The path to the array inside the Zarr store. Example: "/groupA/groupB/outputs/my-array"
        chunks : list[VirtualChunkSpec],
            The list of virtual chunks to add
        validate_containers: bool
            If set to true, ignore virtual references for locations that don't match any existing virtual chunk container


        Returns
        -------
        list[tuple[int, ...]] | None

            If all virtual references where successfully updated, it returns None.
            If there were validation errors, it returns the chunk indices of all failed references.
        """
        return await self._store.set_virtual_refs_async(
            array_path, chunks, validate_containers
        )

    async def delete(self, key: str) -> None:
        """Remove a key from the store

        Parameters
        ----------
        key : str
        """
        return await self._store.delete(key)

    async def delete_dir(self, prefix: str) -> None:
        """Delete a prefix

        Parameters
        ----------
        prefix : str
        """
        return await self._store.delete_dir(prefix)

    @property
    def supports_partial_writes(self) -> bool:
        """Does the store support partial writes?"""
        return self._store.supports_partial_writes

    async def set_partial_values(
        self, key_start_values: Iterable[tuple[str, int, BytesLike]]
    ) -> None:
        """Store values at a given key, starting at byte range_start.

        Parameters
        ----------
        key_start_values : list[tuple[str, int, BytesLike]]
            set of key, range_start, values triples, a key may occur multiple times with different
            range_starts, range_starts (considering the length of the respective values) must not
            specify overlapping ranges for the same key
        """
        # NOTE: pyo3 does not implicit conversion from an Iterable to a rust iterable. So we convert it
        # to a list here first. Possible opportunity for optimization.
        # NOTE: currently we only implement the case where the values are bytes
        return await self._store.set_partial_values(list(key_start_values))  # type: ignore[arg-type]

    @property
    def supports_listing(self) -> bool:
        """Does the store support listing?"""
        return self._store.supports_listing

    @property
    def supports_consolidated_metadata(self) -> bool:
        return self._store.supports_consolidated_metadata

    @property
    def supports_deletes(self) -> bool:
        return self._store.supports_deletes

    def list(self) -> AsyncIterator[str]:
        """Retrieve all keys in the store.

        Returns
        -------
        AsyncIterator[str, None]
        """
        # This method should be async, like overridden methods in child classes.
        # However, that's not straightforward:
        # https://stackoverflow.com/questions/68905848

        # The zarr spec specefies that that this and other
        # listing methods should not be async, so we need to
        # wrap the async method in a sync method.
        return self._store.list()

    def list_prefix(self, prefix: str) -> AsyncIterator[str]:
        """Retrieve all keys in the store that begin with a given prefix. Keys are returned relative
        to the root of the store.

        Parameters
        ----------
        prefix : str

        Returns
        -------
        AsyncIterator[str, None]
        """
        # The zarr spec specefies that that this and other
        # listing methods should not be async, so we need to
        # wrap the async method in a sync method.
        return self._store.list_prefix(prefix)

    def list_dir(self, prefix: str) -> AsyncIterator[str]:
        """
        Retrieve all keys and prefixes with a given prefix and which do not contain the character
        “/” after the given prefix.

        Parameters
        ----------
        prefix : str

        Returns
        -------
        AsyncIterator[str, None]
        """
        # The zarr spec specefies that that this and other
        # listing methods should not be async, so we need to
        # wrap the async method in a sync method.
        return self._store.list_dir(prefix)

    async def getsize(self, key: str) -> int:
        return await self._store.getsize(key)

    async def getsize_prefix(self, prefix: str) -> int:
        return await self._store.getsize_prefix(prefix)



"""
This scripts is the subject of a post in the Earthmover blog.

It uses Icechunk to simulate a series of concurrent bank transactions, and
explores different approaches to conflict resolution.
"""

import multiprocessing
import random
import time
from collections.abc import Callable
from concurrent.futures import ProcessPoolExecutor
from datetime import UTC, datetime
from enum import Enum

import icechunk
import zarr

MAX_INITIAL_BALANCE = 100_000  # cents


class WireResult(Enum):
    DONE = 1
    NOT_ENOUGH_BALANCE = 2


def create_repo(num_accounts: int) -> icechunk.Repository:
    storage = icechunk.s3_storage(
        region="us-east-1",
        endpoint_url="http://localhost:9000",
        allow_http=True,
        force_path_style=True,
        bucket="testbucket",
        prefix="bank_accounts_example_" + str(int(time.time() * 1000)),
        access_key_id="minio123",
        secret_access_key="minio123",
    )
    repo = icechunk.Repository.create(storage=storage)

    session = repo.writable_session("main")
    group = zarr.group(store=session.store, overwrite=True)
    group.create_array(
        "accounts",
        shape=(num_accounts,),
        chunks=(1,),
        dtype="uint64",  # cents
        fill_value=0,
        compressors=None,
    )
    session.commit("Accounts array created")
    return repo


def initial_balances(max_initial_balance: int, repo: icechunk.Repository) -> int:
    """Add initial random balance to every account.

    Make a commit when done.
    """

    session = repo.writable_session("main")
    group = zarr.open_group(store=session.store, mode="r")
    array = group["accounts"]
    num_accounts = array.shape[0]
    for account in range(num_accounts):
        balance = random.randint(0, max_initial_balance + 1)
        array[account] = balance
    total_balance = sum(array)
    session.commit("Account balances initialized")
    return total_balance


def transfer(
    from_account: int, to_account: int, amount: int, array: zarr.Array
) -> WireResult:
    """Transfer amount cents from from_account to to_account.

    It records the operation in the array but it doesn't attempt to commit the session.
    """

    # pretend we are doing some actual work
    time.sleep(random.uniform(0.5, 1))

    balance = array[from_account]
    res = WireResult.DONE
    if balance < amount:
        res = WireResult.NOT_ENOUGH_BALANCE
    else:
        array[from_account] = balance - amount
        array[to_account] = array[to_account] + amount
    return res


def unsafe_transfer_task(repo: icechunk.Repository) -> WireResult:
    """Naif approach to transfers.num_accounts.

    This approach fails by generating commit conflicts.
    """

    session = repo.writable_session("main")
    group = zarr.open_group(store=session.store, mode="r")
    array = group["accounts"]
    num_accounts = array.shape[0]
    from_account = random.randint(0, num_accounts - 1)
    to_account = random.randint(0, num_accounts - 1)
    amount = random.randint(0, int(MAX_INITIAL_BALANCE / 5))
    res = transfer(from_account, to_account, amount, array)
    if res == WireResult.DONE:
        session.commit(f"wired ${amount}: {from_account} -> {to_account}")
    return res


def slow_transfer_task(repo: icechunk.Repository) -> WireResult:
    """This approach to concurrent transfers is safe but very slow.

    It simply retries transactions when they conflict.
    """
    session = repo.readonly_session("main")
    group = zarr.open_group(store=session.store, mode="r")
    array = group["accounts"]
    num_accounts = array.shape[0]
    from_account = random.randint(0, num_accounts - 1)
    to_account = random.randint(0, num_accounts - 1)
    amount = random.randint(0, int(MAX_INITIAL_BALANCE / 5))

    while True:
        session = repo.writable_session("main")
        group = zarr.open_group(store=session.store, mode="r")
        array = group["accounts"]
        res = transfer(from_account, to_account, amount, array)
        if res == WireResult.NOT_ENOUGH_BALANCE:
            return res
        if res == WireResult.DONE:
            try:
                session.commit(f"wired ${amount}: {from_account} -> {to_account}")
                return res
            except icechunk.ConflictError:
                pass
    return res


def rebase_transfer_task(repo: icechunk.Repository) -> WireResult:
    """Safe and fast approach to concurrent transfers.

    Rebases conflicting transactions. Retries transactions that cannot be rebased.
    """
    session = repo.readonly_session("main")
    group = zarr.open_group(store=session.store, mode="r")
    array = group["accounts"]
    num_accounts = array.shape[0]
    from_account = random.randint(0, num_accounts - 1)
    to_account = random.randint(0, num_accounts - 1)
    amount = random.randint(0, int(MAX_INITIAL_BALANCE / 5))

    while True:
        session = repo.writable_session("main")
        group = zarr.open_group(store=session.store, mode="r")
        array = group["accounts"]
        res = transfer(from_account, to_account, amount, array)
        if res == WireResult.NOT_ENOUGH_BALANCE:
            return res
        if res == WireResult.DONE:
            try:
                session.commit(
                    f"wired ${amount}: {from_account} -> {to_account}",
                    rebase_with=icechunk.ConflictDetector(),
                )
                return WireResult.DONE
            except icechunk.RebaseFailedError:
                pass


def calculate_total_balance(repo: icechunk.Repository) -> int:
    session = repo.readonly_session(branch="main")
    store = session.store
    group = zarr.open_group(store=store, mode="r")
    array = group["accounts"]
    assert all(balance >= 0 for balance in array)
    return sum(array)


def concurrent_transfers(
    num_accounts: int,
    num_transfers: int,
    total_balance: int,
    transfer_task: Callable[[icechunk.Repository], WireResult],
    repo: icechunk.Repository,
) -> None:
    """Launch concurrent transfers using multi processing.

    When done, it walidates invariants of the system and runs GC.
    """

    # Execute the transfers
    results = []
    with ProcessPoolExecutor(max_workers=10) as exec:
        for res in exec.map(transfer_task, [repo] * num_transfers):
            results.append(res)
            print(f"Done: {len(results)}/{num_transfers}")

    # Verify the end result is valid
    assert len(results) == num_transfers
    succeeded = sum(1 for res in results if res != WireResult.NOT_ENOUGH_BALANCE)
    failed = sum(1 for res in results if res == WireResult.NOT_ENOUGH_BALANCE)
    assert succeeded + failed == num_transfers
    print(f"Succeeded: {succeeded}. Failed: {failed}")

    assert calculate_total_balance(repo) == total_balance
    num_commits = sum(1 for _ in repo.ancestry(branch="main"))
    assert num_commits == succeeded + 3

    # Garbage collect to delete any leftovers
    print(repo.garbage_collect(datetime.now(UTC)))


def main() -> None:
    multiprocessing.set_start_method("forkserver")
    num_accounts = 50
    num_transfers = 50

    repo = create_repo(num_accounts)
    total_balance = initial_balances(MAX_INITIAL_BALANCE, repo)

    ## uncomment to try the different approaches

    # concurrent_transfers(
    #     num_accounts, num_transfers, total_balance, unsafe_transfer_task, repo
    # )
    # concurrent_transfers(
    #     num_accounts, num_transfers, total_balance, slow_transfer_task, repo
    # )
    concurrent_transfers(
        num_accounts, num_transfers, total_balance, rebase_transfer_task, repo
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# pytest -s test_benchmark_writes.py::test_write_chunks
import numpy as np
import pytest

import zarr
from benchmarks import lib
from benchmarks.helpers import get_splitting_config, repo_config_with
from benchmarks.tasks import Executor, write
from icechunk import (
    Repository,
    RepositoryConfig,
    Session,
    VirtualChunkSpec,
    local_filesystem_storage,
)

NUM_CHUNK_REFS = 10_000
NUM_VIRTUAL_CHUNK_REFS = 100_000


pytestmark = pytest.mark.write_benchmark


@pytest.fixture(
    params=[
        pytest.param(None, id="no-splitting"),
        pytest.param(10000, id="split-size-10_000"),
    ]
)
def splitting(request):
    if request.param is None:
        return None
    else:
        try:
            return get_splitting_config(split_size=request.param)
        except ImportError:
            pytest.skip("splitting not supported on this version")


# FIXME: figure out a reasonable default
@pytest.mark.parametrize(
    "executor",
    [
        Executor.threads,
        pytest.param(
            Executor.processes, marks=pytest.mark.skipif(True, reason="hangs on Coiled")
        ),
    ],
)
def test_write_chunks_with_tasks(synth_write_dataset, benchmark, executor):
    """
    Writes chunks orchestrated using 'bare' tasks and executed using
    either a ThreadPoolExecutor or ProcessPoolExecutor.

    Importantly this benchmarks captures timings PER write task, summarizes them,
    and then records them in the .json file.
    """
    timer = benchmark.pedantic(
        # TODO: parametrize over some of these
        write,
        kwargs=dict(
            storage=synth_write_dataset.storage,
            num_arrays=synth_write_dataset.num_arrays,
            shape=synth_write_dataset.shape,
            chunks=synth_write_dataset.chunks,
            task_nchunks=1,
            executor=executor,
            workers=4,
            threads_per_worker=None,
        ),
        iterations=1,
        rounds=1,
    )
    diags = timer.as_dict()
    diags["write_task_raw"] = diags["write_task"]
    diags["write_task"] = lib.stats(np.array(diags["write_task"]))
    benchmark.extra_info["data"] = diags


def test_write_simple_1d(benchmark, simple_write_dataset):
    """Simple write benchmarks. Shows 2-3X slower on GCS compared to S3"""
    dataset = simple_write_dataset
    repo = dataset.create()
    session = repo.writable_session(branch="main")
    store = session.store
    group = zarr.open_group(store, zarr_format=3)
    group.create_array(
        "array",
        shape=dataset.shape,
        chunks=dataset.chunks,
        dtype="int8",
        fill_value=0,
        compressors=None,
    )
    session.commit("initialize")

    def write_task():
        session = repo.writable_session(branch="main")
        array = zarr.open_array(session.store, path="array")
        array[:] = 42
        session.commit("foo")

    benchmark(write_task)


@pytest.mark.benchmark(group="refs-write")
@pytest.mark.parametrize("commit", [True, False])
@pytest.mark.parametrize(
    "repo_config",
    [
        pytest.param(repo_config_with(), id="default-inlined"),
        pytest.param(repo_config_with(inline_chunk_threshold_bytes=0), id="not-inlined"),
    ],
)
def test_write_many_chunk_refs(
    commit: bool, benchmark, repo_config: RepositoryConfig, tmpdir
) -> None:
    """Benchmarking the writing of many chunk refs, inlined and not; committed and not;"""

    # 1. benchmark only the writes with no commit
    # 2. one with commit
    def write_chunk_refs(repo) -> None:
        session = repo.writable_session("main")
        array = zarr.open_array(path="array", store=session.store)
        # TODO: configurable?
        with zarr.config.set({"async.concurrency": 64}):
            array[:] = -1
        if commit:
            session.commit("written!")

    assert repo_config is not None
    repo = Repository.create(storage=local_filesystem_storage(tmpdir), config=repo_config)
    session = repo.writable_session("main")
    group = zarr.group(session.store)
    kwargs = dict(
        name="array",
        shape=(NUM_CHUNK_REFS,),
        chunks=(1,),
        dtype=np.int8,
        dimension_names=("t",),
    )
    try:
        group.create_array(**kwargs)
    except AttributeError:
        group.array(**kwargs)
    session.commit("initialized")

    benchmark(write_chunk_refs, repo)


@pytest.mark.benchmark(group="refs-write")
def test_set_many_virtual_chunk_refs(benchmark, repo) -> None:
    """Benchmark the setting of many virtual chunk refs."""
    session = repo.writable_session("main")
    store = session.store
    group = zarr.group(store)
    kwargs = dict(
        name="array",
        shape=(NUM_VIRTUAL_CHUNK_REFS,),
        chunks=(1,),
        dtype=np.int8,
        dimension_names=("t",),
        overwrite=True,
    )
    try:
        group.create_array(**kwargs)
    except AttributeError:
        group.array(**kwargs)
    session.commit("initialized")

    @benchmark
    def write():
        # always create a new session so we have a clean changelog
        session = repo.writable_session("main")
        store = session.store
        for i in range(NUM_VIRTUAL_CHUNK_REFS):
            store.set_virtual_ref(
                f"array/c/{i}", location=f"s3://foo/bar/{i}.nc", offset=0, length=1
            )


@pytest.mark.benchmark(group="refs-write")
def test_write_split_manifest_refs_full_rewrite(
    benchmark, splitting, large_write_dataset
) -> None:
    dataset = large_write_dataset
    config = repo_config_with(splitting=splitting)
    assert config is not None
    if hasattr(config.manifest, "splitting"):
        assert config.manifest.splitting == splitting
    repo = dataset.create(config=config)
    session = repo.writable_session(branch="main")
    store = session.store
    group = zarr.open_group(store, zarr_format=3)
    group.create_array(
        "array",
        shape=dataset.shape,
        chunks=dataset.chunks,
        dtype="int8",
        fill_value=0,
        compressors=None,
    )
    session.commit("initialize")

    def write_refs() -> Session:
        session = repo.writable_session(branch="main")
        num_chunks = dataset.shape[0] // dataset.chunks[0]
        chunks = [
            VirtualChunkSpec(
                index=[i], location=f"s3://foo/bar/{i}.nc", offset=0, length=1
            )
            for i in range(num_chunks)
        ]

        session.store.set_virtual_refs("array", chunks)
        # (args, kwargs)
        return ((session,), {})

    def commit(session_from_setup):
        session_from_setup.commit("wrote refs")

    benchmark.pedantic(commit, setup=write_refs, iterations=1, rounds=10)


@pytest.mark.benchmark(group="refs-write")
def test_write_split_manifest_refs_append(
    benchmark, splitting, large_write_dataset
) -> None:
    dataset = large_write_dataset
    config = repo_config_with(splitting=splitting)
    assert config is not None
    if hasattr(config.manifest, "splitting"):
        assert config.manifest.splitting == splitting
    repo = dataset.create(config=config)
    session = repo.writable_session(branch="main")
    store = session.store
    group = zarr.open_group(store, zarr_format=3)
    group.create_array(
        "array",
        shape=dataset.shape,
        chunks=dataset.chunks,
        dtype="int8",
        fill_value=0,
        compressors=None,
    )
    session.commit("initialize")

    # yuck, but I'm abusing `rounds` to do a loop and time _only_ the commit.
    global counter
    counter = 0
    rounds = 10
    num_chunks = dataset.shape[0] // dataset.chunks[0]
    batch_size = num_chunks // rounds

    def write_refs() -> Session:
        global counter
        session = repo.writable_session(branch="main")
        chunks = [
            VirtualChunkSpec(
                index=[i], location=f"s3://foo/bar/{i}.nc", offset=0, length=1
            )
            for i in range(counter * batch_size, counter * batch_size + batch_size)
        ]
        counter += 1
        session.store.set_virtual_refs("array", chunks)
        # (args, kwargs)
        return ((session,), {})

    def commit(session_from_setup):
        session_from_setup.commit("wrote refs")

    benchmark.pedantic(commit, setup=write_refs, iterations=1, rounds=rounds)



# Consolidate some "big" dataset creation
# My workflow is:
#   1. just create-deepak-env v0.1.0a12
#   2. conda activate icechunk-v0.1.0a12
#   3. python benchmarks/create-era5.py

import argparse
import datetime
import math
import random
from enum import StrEnum, auto
from typing import Any

import humanize
import pandas as pd
from packaging.version import Version

import dask
import icechunk as ic
import xarray as xr
import zarr
from benchmarks import helpers
from benchmarks.datasets import Dataset, IngestDataset
from dask.diagnostics import ProgressBar
from icechunk.xarray import to_icechunk

logger = helpers.setup_logger()

ICECHUNK_FORMAT = f"v{ic.spec_version():d}"
ZARR_KWARGS = dict(zarr_format=3, consolidated=False)


class Mode(StrEnum):
    APPEND = auto()
    CREATE = auto()
    OVERWRITE = auto()
    VERIFY = auto()


ERA5_WB = IngestDataset(
    name="ERA5-WB",
    prefix="era5_weatherbench2",
    source_uri="gs://weatherbench2/datasets/era5/1959-2023_01_10-full_37-1h-0p25deg-chunk-1.zarr",
    engine="zarr",
    read_chunks={"time": 24 * 3, "level": 1},
    write_chunks={"time": 1, "level": 1, "latitude": 721, "longitude": 1440},
    group="1x721x1440",
    arrays=["2m_temperature", "10m_u_component_of_wind", "10m_v_component_of_wind"],
)


def verify(dataset: Dataset, *, ingest: IngestDataset, seed: int | None = None):
    random.seed(seed)

    format = ICECHUNK_FORMAT
    prefix = f"{format}/"
    dataset.storage_config = dataset.storage_config.with_extra(prefix=prefix)
    repo = ic.Repository.open(dataset.storage)
    session = repo.readonly_session(branch="main")
    instore = xr.open_dataset(
        session.store, group=dataset.group, engine="zarr", chunks=None, consolidated=False
    )
    time = pd.Timestamp(random.choice(instore.time.data.tolist()))
    logger.info(f"Verifying {ingest.name} for {seed=!r}, {time=!r}")
    actual = instore.sel(time=time)

    ds = ingest.open_dataset(chunks=None)
    expected = ds[list(instore.data_vars)].sel(time=time)

    # I add global attrs, don't compare those
    expected.attrs.clear()
    actual.attrs.clear()

    # TODO: Parallelize the compare in `assert_identical` upstream
    with ProgressBar():
        actual, expected = dask.compute(actual, expected)
        # assert (expected == actual).all().to_array().all()
    xr.testing.assert_identical(expected, actual)
    logger.info("Successfully verified!")


def write(
    dataset: Dataset,
    *,
    ingest: IngestDataset,
    mode: Mode,
    nyears: int | None = None,
    arrays_to_write: list[str] | None = None,
    extra_attrs: dict[str, str] | None = None,
    initialize_all_vars: bool = False,
    dry_run: bool = False,
) -> None:
    """
    dataset: Dataset t write
    arrays_to_write: list[str],
    extra_attrs: any attributes to add
    initialize_all_vars: whether to write all coordinate arrays, and metadata for ALL data_vars.

    Usually, initialize_all_vars=True for benchmarks, but not for the "public dataset".
    For benchmarks,
      1. We write all the metadata and coordinate arrays to make a "big" snapshot.
      2. We only write a few arrays to save time.
    """
    import coiled

    import distributed

    SELECTOR = {"time": slice(nyears * 365 * 24) if nyears is not None else slice(None)}
    if mode in [Mode.CREATE, Mode.OVERWRITE]:
        write_mode = "w"
    elif mode is Mode.APPEND:
        write_mode = "a"

    ic_kwargs = dict(group=dataset.group, mode=write_mode)

    ds = ingest.open_dataset()
    if arrays_to_write is not None:
        towrite = ds[arrays_to_write].isel(SELECTOR)
    else:
        towrite = ds.isel(SELECTOR)
    towrite.attrs["selector"] = str(SELECTOR)
    towrite.attrs.update(extra_attrs or {})
    for v in towrite:
        towrite[v].encoding["chunks"] = tuple(
            ingest.write_chunks[dim] for dim in ds[v].dims
        )

    nchunks = tuple(
        math.prod((var.sizes[dim] // ingest.write_chunks[dim] + 1) for dim in var.dims)
        for _, var in towrite.data_vars.items()
    )
    logger.info(
        f"Size: {humanize.naturalsize(towrite.nbytes)}, "
        f"Total nchunks= {humanize.intcomma(sum(nchunks))}, "
        f"per array: {[humanize.intcomma(i) for i in nchunks]}"
    )

    repo = ic.Repository.open(dataset.storage)

    if dry_run:
        print("Dry run. Exiting")
        return

    ckwargs = dataset.storage_config.get_coiled_kwargs()
    session = repo.writable_session("main")
    with coiled.Cluster(
        name=f"icechunk-ingest-{ICECHUNK_FORMAT}-{ingest.name}",
        shutdown_on_close=False,
        n_workers=(4, 200),
        worker_cpu=2,
        workspace=ckwargs["workspace"],
        region=ckwargs["region"],
    ) as cluster:
        # https://docs.coiled.io/user_guide/clusters/environ.html
        cluster.send_private_envs(dataset.storage_config.env_vars)
        client = distributed.Client(cluster)  # type: ignore[no-untyped-call]
        print(client)
        with distributed.performance_report(  # type: ignore[no-untyped-call]
            f"reports/{ingest.name}-ingest-{dataset.storage_config.store}-{ICECHUNK_FORMAT}-{datetime.datetime.now()}.html"
        ):
            logger.info(f"Started writing {tuple(towrite.data_vars)}.")
            with zarr.config.set({"async.concurrency": 24}):
                to_icechunk(towrite, session=session, **ic_kwargs, split_every=32)
    session.commit("ingest!")
    logger.info(f"Finished writing {tuple(towrite.data_vars)}.")


def setup_dataset(
    dataset: Dataset,
    *,
    ingest: IngestDataset,
    mode: Mode,
    dry_run: bool = False,
    **kwargs: Any,
) -> None:
    # commit = helpers.get_commit(ref)
    format = ICECHUNK_FORMAT
    logger.info(f"Writing {ingest.name} for {format}, {kwargs=}")
    prefix = f"{format}/"
    dataset.storage_config = dataset.storage_config.with_extra(prefix=prefix)
    if mode is Mode.CREATE:
        logger.info("Creating new repository")
        repo = dataset.create(clear=True)
        logger.info("Initializing root group")
        session = repo.writable_session("main")
        zarr.open_group(session.store, mode="w-")
        session.commit("initialized root group")
        logger.info("Initialized root group")

    write(
        dataset,
        ingest=ingest,
        mode=mode,
        initialize_all_vars=False,
        dry_run=dry_run,
        **kwargs,
    )


def get_version() -> str:
    version = Version(ic.__version__)
    if version.pre is not None and "a" in version.pre:
        return f"icechunk-v{version.base_version}-alpha.{version.pre[1]}"
    else:
        return f"icechunk-v{version.base_version}"


if __name__ == "__main__":
    helpers.assert_cwd_is_icechunk_python()

    parser = argparse.ArgumentParser()
    parser.add_argument("store", help="object store to write to")
    parser.add_argument(
        "--mode", help="'create'/'overwrite'/'append'/'verify'", default="append"
    )
    parser.add_argument(
        "--nyears", help="number of years to write (from start)", default=None, type=int
    )
    parser.add_argument("--dry-run", action="store_true", help="dry run?", default=False)
    parser.add_argument(
        "--append", action="store_true", help="append or create?", default=False
    )
    parser.add_argument("--arrays", help="arrays to write", nargs="+", default=[])
    parser.add_argument("--seed", help="random seed for verify", default=None, type=int)
    parser.add_argument(
        "--debug", help="write to debug bucket?", default=False, action="store_true"
    )

    args = parser.parse_args()
    if args.mode == "create":
        mode = Mode.CREATE
    elif args.mode == "append":
        mode = Mode.APPEND
    elif args.mode == "overwrite":
        mode = Mode.OVERWRITE
    elif args.mode == "verify":
        mode = Mode.VERIFY
    else:
        raise ValueError(
            f"mode must be one of ['create', 'overwrite', 'append', 'verify']. Received {args.mode=!r}"
        )

    ingest = ERA5_WB
    dataset = ingest.make_dataset(store=args.store, debug=args.debug)
    logger.info(ingest)
    logger.info(dataset)
    logger.info(args)
    ds = ingest.open_dataset()
    if mode is Mode.VERIFY:
        verify(dataset, ingest=ingest, seed=args.seed)
    else:
        setup_dataset(
            dataset,
            ingest=ingest,
            nyears=args.nyears,
            mode=mode,
            arrays_to_write=args.arrays or ingest.arrays,
            dry_run=args.dry_run,
        )



# module

from typing import TypeAlias

from icechunk._icechunk_python import (
    AzureCredentials,
    AzureStaticCredentials,
    BasicConflictSolver,
    CachingConfig,
    CompressionAlgorithm,
    CompressionConfig,
    Conflict,
    ConflictDetector,
    ConflictError,
    ConflictSolver,
    ConflictType,
    Credentials,
    Diff,
    GcsBearerCredential,
    GcsCredentials,
    GcsStaticCredentials,
    GCSummary,
    IcechunkError,
    ManifestConfig,
    ManifestFileInfo,
    ManifestPreloadCondition,
    ManifestPreloadConfig,
    ManifestSplitCondition,
    ManifestSplitDimCondition,
    ManifestSplittingConfig,
    ObjectStoreConfig,
    RebaseFailedError,
    RepositoryConfig,
    S3Credentials,
    S3Options,
    S3StaticCredentials,
    SnapshotInfo,
    Storage,
    StorageConcurrencySettings,
    StorageRetriesSettings,
    StorageSettings,
    VersionSelection,
    VirtualChunkContainer,
    VirtualChunkSpec,
    __version__,
    initialize_logs,
    set_logs_filter,
    spec_version,
)
from icechunk.credentials import (
    AnyAzureCredential,
    AnyAzureStaticCredential,
    AnyCredential,
    AnyGcsCredential,
    AnyGcsStaticCredential,
    AnyS3Credential,
    azure_credentials,
    azure_from_env_credentials,
    azure_static_credentials,
    containers_credentials,
    gcs_credentials,
    gcs_from_env_credentials,
    gcs_refreshable_credentials,
    gcs_static_credentials,
    s3_anonymous_credentials,
    s3_credentials,
    s3_from_env_credentials,
    s3_refreshable_credentials,
    s3_static_credentials,
)
from icechunk.repository import Repository
from icechunk.session import ForkSession, Session
from icechunk.storage import (
    AnyObjectStoreConfig,
    azure_storage,
    gcs_storage,
    gcs_store,
    http_store,
    in_memory_storage,
    local_filesystem_storage,
    local_filesystem_store,
    r2_storage,
    s3_storage,
    s3_store,
    tigris_storage,
)
from icechunk.store import IcechunkStore

__all__ = [
    "AnyAzureCredential",
    "AnyAzureStaticCredential",
    "AnyCredential",
    "AnyGcsCredential",
    "AnyGcsStaticCredential",
    "AnyObjectStoreConfig",
    "AnyS3Credential",
    "AzureCredentials",
    "AzureStaticCredentials",
    "BasicConflictSolver",
    "CachingConfig",
    "CompressionAlgorithm",
    "CompressionConfig",
    "Conflict",
    "ConflictDetector",
    "ConflictError",
    "ConflictSolver",
    "ConflictType",
    "Credentials",
    "Diff",
    "ForkSession",
    "GCSummary",
    "GcsBearerCredential",
    "GcsCredentials",
    "GcsStaticCredentials",
    "IcechunkError",
    "IcechunkStore",
    "ManifestConfig",
    "ManifestFileInfo",
    "ManifestPreloadCondition",
    "ManifestPreloadConfig",
    "ManifestSplitCondition",
    "ManifestSplitDimCondition",
    "ManifestSplittingConfig",
    "ObjectStoreConfig",
    "RebaseFailedError",
    "Repository",
    "RepositoryConfig",
    "S3Credentials",
    "S3Options",
    "S3StaticCredentials",
    "Session",
    "SnapshotInfo",
    "Storage",
    "StorageConcurrencySettings",
    "StorageRetriesSettings",
    "StorageSettings",
    "VersionSelection",
    "VirtualChunkContainer",
    "VirtualChunkSpec",
    "__version__",
    "azure_credentials",
    "azure_from_env_credentials",
    "azure_static_credentials",
    "azure_storage",
    "containers_credentials",
    "gcs_credentials",
    "gcs_from_env_credentials",
    "gcs_refreshable_credentials",
    "gcs_static_credentials",
    "gcs_storage",
    "gcs_store",
    "http_store",
    "in_memory_storage",
    "initialize_logs",
    "local_filesystem_storage",
    "local_filesystem_store",
    "print_debug_info",
    "r2_storage",
    "s3_anonymous_credentials",
    "s3_credentials",
    "s3_from_env_credentials",
    "s3_refreshable_credentials",
    "s3_static_credentials",
    "s3_storage",
    "s3_store",
    "set_logs_filter",
    "spec_version",
    "tigris_storage",
]


def print_debug_info() -> None:
    import platform
    from importlib import import_module

    print(f"platform:  {platform.platform()}")
    print(f"python:  {platform.python_version()}")
    print(f"icechunk:  {__version__}")
    for package in ["zarr", "numcodecs", "xarray", "virtualizarr"]:
        try:
            print(f"{package}:  {import_module(package).__version__}")
        except ModuleNotFoundError:
            continue


# This monkey patch is a bit annoying. Python dicts preserve insertion order
# But this gets mapped to a Rust HashMap which does *not* preserve order
# So on the python side, we can accept a dict as a nicer API, and immediately
# convert it to tuples that preserve order, and pass those to Rust

SplitSizesDict: TypeAlias = dict[
    ManifestSplitCondition,
    dict[
        ManifestSplitDimCondition.Axis
        | ManifestSplitDimCondition.DimensionName
        | ManifestSplitDimCondition.Any,
        int,
    ],
]


def from_dict(split_sizes: SplitSizesDict) -> ManifestSplittingConfig:
    unwrapped = tuple((k, tuple(v.items())) for k, v in split_sizes.items())
    return ManifestSplittingConfig(unwrapped)


def to_dict(config: ManifestSplittingConfig) -> SplitSizesDict:
    return {
        split_condition: dict(dim_conditions)
        for split_condition, dim_conditions in config.split_sizes
    }


ManifestSplittingConfig.from_dict = from_dict  # type: ignore[method-assign]
ManifestSplittingConfig.to_dict = to_dict  # type: ignore[attr-defined]

initialize_logs()



#!/usr/bin/env python3
#
# This is just a scratch script for testing purposes
# coiled notebook start --sync --software icechunk-alpha-12 --vm-type m5.4xlarge
import subprocess

# software = "icechunk-alpha-12"
vm_type = {
    "s3": "m5.4xlarge",
    "gcs": None,
    "tigris": None,
}
ref = "icechunk-v0.1.0-alpha.12"

COILED_SOFTWARE = {
    "icechunk-v0.1.0-alpha.1": "icechunk-alpha-release",
    "icechunk-v0.1.0-alpha.12": "icechunk-alpha-12",
}
software = COILED_SOFTWARE[ref]

cmd = f'python benchmarks/runner.py --where coiled --pytest "-k zarr_open" {ref}'
subprocess.run(
    [
        "coiled",
        "run",
        "--name",
        "icebench-712f1eb2",
        "--sync",
        "--sync-ignore='python/ reports/ profiling/'",
        "--keepalive",
        "5m",
        "--workspace=earthmover-devs",
        "--vm-type=m5.4xlarge",
        "--software=icechunk-bench-712f1eb2",
        "--region=us-east-1",
        "pytest -v benchmarks/",
    ]
)



import copy
from typing import cast

import pytest

from benchmarks import helpers
from benchmarks.datasets import (
    LARGE_1D,
    LARGE_MANIFEST_SHARDED,
    LARGE_MANIFEST_UNSHARDED,
    PANCAKE_WRITES,
    SIMPLE_1D,
    TEST_BUCKETS,
    BenchmarkReadDataset,
    BenchmarkWriteDataset,
    Dataset,
)
from icechunk import Repository, local_filesystem_storage

try:
    from icechunk import ManifestSplittingConfig  # noqa: F401

    no_splitting = False
except ImportError:
    no_splitting = True


def request_to_dataset(request, moar_prefix: str = "") -> Dataset:
    extra_prefix = request.config.getoption("--icechunk-prefix") + moar_prefix
    where = request.config.getoption("--where")
    ds = copy.deepcopy(request.param)
    if where == "local" and ds.skip_local:
        pytest.skip()
    # this gets run multiple times because the fixture scope is 'function'
    # so we need this `force_idempotent` ugliness
    ds.storage_config = ds.storage_config.with_overwrite(
        **TEST_BUCKETS[where]
    ).with_extra(prefix=extra_prefix, force_idempotent=True)
    return ds


@pytest.fixture(scope="function")
def repo(tmpdir: str) -> Repository:
    return Repository.create(storage=local_filesystem_storage(tmpdir))


@pytest.fixture(params=[pytest.param(PANCAKE_WRITES, id="pancake-writes")])
def synth_write_dataset(request) -> BenchmarkWriteDataset:
    ds = request_to_dataset(request, moar_prefix=helpers.rdms())
    return cast(BenchmarkWriteDataset, ds)


@pytest.fixture(params=[pytest.param(SIMPLE_1D, id="simple-1d")])
def simple_write_dataset(request) -> BenchmarkWriteDataset:
    ds = request_to_dataset(request, moar_prefix=helpers.rdms())
    return cast(BenchmarkWriteDataset, ds)


@pytest.fixture(params=[pytest.param(LARGE_1D, id="large-1d")])
def large_write_dataset(request) -> BenchmarkWriteDataset:
    moar_prefix = helpers.rdms()
    ds = request_to_dataset(request, moar_prefix=moar_prefix)
    return cast(BenchmarkWriteDataset, ds)


@pytest.fixture(
    params=[
        # pytest.param(GB_8MB_CHUNKS, id="gb-8mb"),
        # pytest.param(GB_128MB_CHUNKS, id="gb-128mb"),
        # pytest.param(ERA5_SINGLE, id="era5-single"),
        # pytest.param(ERA5, id="era5-weatherbench"),
        # pytest.param(ERA5_ARCO, id="era5-arco"),
        pytest.param(LARGE_MANIFEST_UNSHARDED, id="large-manifest-no-split"),
        pytest.param(
            LARGE_MANIFEST_SHARDED,
            id="large-manifest-split",
            marks=pytest.mark.skipif(no_splitting, reason="no splitting"),
        ),
    ],
)
def synth_dataset(request) -> BenchmarkReadDataset:
    """For now, these are synthetic datasets stored in the cloud."""
    ds = request_to_dataset(request)
    if ds.setupfn is None:
        # these datasets aren't automatically set up
        # so skip if the data haven't been written yet.
        try:
            ds.store()
        except ValueError as e:
            pytest.skip(reason=str(e))
    return cast(BenchmarkReadDataset, ds)


# This hook is used instead of `pyproject.toml` so that we can run the benchmark infra
# on versions older than alpha-13
# TODO: Migrate to pyproject.toml after 1.0 has been released.
def pytest_configure(config):
    config.addinivalue_line(
        "markers", "setup_benchmarks: Run synthetic dataset setup code for benchmarks."
    )
    config.addinivalue_line("markers", "read_benchmark: Read benchmark.")
    config.addinivalue_line("markers", "write_benchmark: Write benchmark.")


def pytest_addoption(parser):
    parser.addoption(
        "--icechunk-prefix",
        action="store",
        default="",
        help="""
        extra prefix that is prefix-ed to the configured prefix or
        path in icechunk benchmark datasets.
        """,
    )

    parser.addoption(
        "--force-setup",
        action="store",
        default=True,
        help="""
        Force running the setup_benchmarks code even if there is a valid repo
        for this icechunk version at that URI. True by default.
        """,
    )

    parser.addoption(
        "--where",
        action="store",
        help="Where to run icechunk benchmarks? [local|s3|gcs].",
        default="local",
    )



# Regression test for GH1006
# https://github.com/earth-mover/icechunk/issues/1006

import datetime

import pytest

import dask
import icechunk as ic
import xarray as xr
import zarr
from icechunk.xarray import to_icechunk

p = zarr.core.buffer.default_buffer_prototype()

PLOT = False
SPLIT_EVERY = 128
DIMS = ("x", "y")
IC_STORAGE = ic.local_filesystem_storage(
    f"/tmp/test/icechunk_data_corrupted/{str(datetime.datetime.now()).split(' ')[-1]}",
)


def do_test(scheduler) -> None:
    # Writing the initial dataset
    if scheduler in ["processes", "sync"]:
        CHUNKX = 3
        CHUNKY = 6
        X = CHUNKX * 1 + 1
        Y = CHUNKY * 1 + 1
        dX = 1
        dY = 1
    else:
        CHUNKX = 300
        CHUNKY = 600
        X = CHUNKX * 1 + 1
        Y = CHUNKY * 25 + 1
        dX = 1
        dY = 1

    def plot() -> None:
        if not PLOT:
            return
        import matplotlib.pyplot as plt

        # kwargs = dict(ylim=(X - CHUNKX, X + dX + 1), xlim=(Y - CHUNKY // 10, Y + dY + 1))
        plt.pcolor(
            xr.open_zarr(session.store, consolidated=False)["a"].isel(
                x=slice(X - CHUNKX, None), y=slice(Y - CHUNKY, None)
            ),
            vmin=0,
            vmax=4,
        )
        # plt.gca().set_yticks(np.arange(0, X + dX + 1, CHUNKX))
        # plt.gca().set_xticks(np.arange(0, Y + dY + 1, CHUNKY))
        plt.xlim(0, Y + 1)
        plt.ylim(0, X + 1)
        plt.axhline(CHUNKX - 1, color="g")
        plt.axvline(CHUNKY - 1, color="g")
        plt.axhline(X - 1, color="r")
        plt.axvline(Y - 1, color="r")
        plt.grid(True)

    initial = xr.Dataset(
        {
            "a": xr.DataArray(
                1.1, dims=DIMS, coords={"x": list(range(X)), "y": list(range(Y))}
            )
        }
    ).chunk(x=CHUNKX, y=CHUNKY)

    repo = ic.Repository.open_or_create(IC_STORAGE)
    session = repo.writable_session("main")
    with dask.config.set(scheduler=scheduler):
        to_icechunk(initial, session=session, mode="w", split_every=SPLIT_EVERY)
    plot()
    session.commit("initial write")
    assert (xr.open_zarr(session.store, consolidated=False)["a"] >= 1).all()

    print("finished initial write")
    print("-======================")
    print("appending along x")

    # Appending data on the X axis
    x_append_data = xr.Dataset(
        {
            "a": xr.DataArray(
                2.2,
                dims=DIMS,
                coords={"x": list(range(X, X + dX)), "y": list(range(Y))},
            )
        }
    ).chunk(x=dX, y=CHUNKY)
    session = repo.writable_session("main")
    with dask.config.set(scheduler=scheduler):
        to_icechunk(
            x_append_data, session=session, append_dim="x", split_every=SPLIT_EVERY
        )
    plot()
    session.commit("append along x")
    assert (xr.open_zarr(session.store, consolidated=False)["a"] >= 1).all()

    print("-======================")
    print("appending along y")

    # Appending data on the Y axis, this part is the one that corrupts the data
    y_append_data = xr.Dataset(
        {
            "a": xr.DataArray(
                3.3,
                dims=DIMS,
                coords={"x": list(range(X + dX)), "y": list(range(Y, Y + dY))},
            ),
        }
    ).chunk(x=CHUNKX, y=-1)
    session = repo.writable_session("main")
    with dask.config.set(scheduler=scheduler):
        to_icechunk(
            y_append_data, session=session, append_dim="y", split_every=SPLIT_EVERY
        )
    plot()
    session.commit("append along y")
    assert (xr.open_zarr(session.store, consolidated=False)["a"] >= 1).all()


@pytest.mark.parametrize("scheduler", ["threads", "processes"])
def test_dask_distributed_appends(scheduler) -> None:
    do_test(scheduler)


if __name__ == "__main__":
    do_test("threads")
    do_test("processes")
